import strax
import numpy as np
from immutabledict import immutabledict

export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option(
        'area_vs_width_nbins',
        type=int, default=60,
        help='Number of bins for area vs width histogram for online monitor. '
             'NB: this is a 2D histogram'),
    strax.Option(
        'area_vs_width_bounds',
        type=tuple, default=((0, 5), (0, 5)),
        help='Boundaries of log-log histogram of area vs width'),
    strax.Option(
        'area_vs_width_cut_string',
        type=str, default='',
        help='Selection (like selection_str) applied to data for '
             '"area_vs_width_hist_clean", cuts should be separated using "&"'
             'For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)'
             'Default is no selection (other than "area_vs_width_min_gap")'),
    strax.Option(
        'lone_hits_area_bounds',
        type=tuple, default=(0, 1500),
        help='Boundaries area histogram of lone hits [ADC]'),
    strax.Option(
        'online_peak_monitor_nbins',
        type=int, default=100,
        help='Number of bins of histogram of online monitor. Will be used '
             'for: '
             'lone_hits_area-histogram, '
             'area_fraction_top-histogram, '
             'online_se_gain estimate (histogram is not stored), '
    ),
    strax.Option(
        'lone_hits_cut_string',
        type=str,
        default='(area >= 50) & (area <= 250)',
        help='Selection (like selection_str) applied to data for '
             '"lone-hits", cuts should be separated using "&")'),
    strax.Option(
        'lone_hits_min_gap',
        type=int,
        default=15_000,
        help='Minimal gap [ns] between consecutive lone-hits. To turn off '
             'this cut, set to 0.'),
    strax.Option(
        'n_tpc_pmts', type=int,
        help='Number of TPC PMTs'),
    strax.Option(
        'online_se_bounds',
        type=tuple, default=(7, 70),
        help='Window for online monitor [PE] to look for the SE gain, value'
    )
)
class OnlinePeakMonitor(strax.Plugin):
    """
    Plugin to write data to the online-monitor. Data that is written by
    this plugin should be small such as to not overload the runs-
    database.

    This plugin takes 'peak_basics' and 'lone_hits'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'online_peak_monitor' with info on the lone-hits and peaks
    """
    depends_on = ('peak_basics', 'lone_hits')
    provides = 'online_peak_monitor'
    data_kind = 'online_peak_monitor'
    __version__ = '0.0.5'
    rechunk_on_save = False

    def infer_dtype(self):
        n_bins_area_width = self.config['area_vs_width_nbins']
        bounds_area_width = self.config['area_vs_width_bounds']

        n_bins = self.config['online_peak_monitor_nbins']

        n_tpc_pmts = self.config['n_tpc_pmts']
        dtype = [
            (('Start time of the chunk', 'time'),
             np.int64),
            (('End time of the chunk', 'endtime'),
             np.int64),
            (('Area vs width histogram (log-log)', 'area_vs_width_hist'),
             (np.int64, (n_bins_area_width, n_bins_area_width))),
            (('Area vs width edges (log-space)', 'area_vs_width_bounds'),
             (np.float64, np.shape(bounds_area_width))),
            (('Lone hits areas histogram [ADC-counts]', 'lone_hits_area_hist'),
             (np.int64, n_bins)),
            (('Lone hits areas bounds [ADC-counts]', 'lone_hits_area_bounds'),
             (np.float64, 2)),
            (('Lone hits per channel', 'lone_hits_per_channel'),
             (np.int64, n_tpc_pmts)),
            (('AFT histogram', 'aft_hist'),
             (np.int64, n_bins)),
            (('AFT bounds', 'aft_bounds'),
             (np.float64, 2)),
            (('Number of contributing channels histogram', 'n_channel_hist'),
             (np.int64, n_tpc_pmts)),
            (('Single electron gain', 'online_se_gain'),
             np.float32),
        ]
        return dtype

    def compute(self, peaks, lone_hits, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end
        n_pmt = self.config['n_tpc_pmts']
        n_bins = self.config['online_peak_monitor_nbins']

        # Bounds for histograms
        res['area_vs_width_bounds'] = self.config['area_vs_width_bounds']
        res['lone_hits_area_bounds'] = self.config['lone_hits_area_bounds']

        # -- Peak vs area 2D histogram --
        # Always cut out unphysical peaks
        sel = (peaks['area'] > 0) & (peaks['range_50p_area'] > 0)
        res['area_vs_width_hist'] = self.area_width_hist(peaks[sel])
        del sel

        # -- Lone hit properties --
        # Make a mask with the cuts.
        # Now only take lone hits that are separated in time.
        if len(lone_hits):
            lh_timedelta = lone_hits[1:]['time'] - strax.endtime(lone_hits)[:-1]
            # Hits on the left are far away? (assume first is because of chunk bound)
            mask = np.hstack([True, lh_timedelta > self.config['lone_hits_min_gap']])
            # Hits on the right are far away? (assume last is because of chunk bound)
            mask &= np.hstack([lh_timedelta > self.config['lone_hits_min_gap'], True])
        else:
            mask = []
        masked_lh = strax.apply_selection(lone_hits[mask],
                                          selection_str=self.config['lone_hits_cut_string'])

        # Make histogram of ADC counts
        # NB: LONE HITS AREA ARE IN ADC!
        lone_hit_areas, _ = np.histogram(masked_lh['area'],
                                         bins=n_bins,
                                         range=self.config['lone_hits_area_bounds'])

        lone_hit_channel_count, _ = np.histogram(masked_lh['channel'],
                                                 bins=n_pmt,
                                                 range=[0, n_pmt])
        # Count number of lone-hits per PMT
        res['lone_hits_area_hist'] = lone_hit_areas
        res['lone_hits_per_channel'] = lone_hit_channel_count
        # Clear mask, don't re-use
        del mask

        # -- AFT histogram --
        aft_b = [0, 1]
        aft_hist, _ = np.histogram(peaks['area_fraction_top'], bins=n_bins, range=aft_b)
        res['aft_hist'] = aft_hist
        res['aft_bounds'] = aft_b

        # Estimate Single Electron (SE) gain
        se_hist, se_bins = np.histogram(peaks['area'], bins=n_bins,
                                        range=self.config['online_se_bounds'])
        bin_centers = (se_bins[1:] + se_bins[:-1]) / 2
        res['online_se_gain'] = bin_centers[np.argmax(se_hist)]
        return res

    def area_width_hist(self, data):
        """Make area vs width 2D-hist"""
        hist, _, _ = np.histogram2d(
            np.log10(data['area']),
            np.log10(data['range_50p_area']),
            range=self.config['area_vs_width_bounds'],
            bins=self.config['area_vs_width_nbins'])
        return hist.T


@export
@strax.takes_config(
    strax.Option(
        'channel_map', 
        track=False, 
        type=immutabledict,
        help='immutabledict mapping subdetector to (min, max) '
             'channel number.'),
    strax.Option(
        'events_area_bounds',
        type=tuple, default=(-0.5, 130.5),
        help='Boundaries area histogram of events_nv_area_per_chunk [PE]'),
    strax.Option(
        'events_area_nbins',
        type=int, default=131,
        help='Number of bins of histogram of events_nv_area_per_chunk, '
             'defined value 1 PE/bin')
)
class OnlineMonitorNV(strax.Plugin):
    """
    Plugin to write data of nVeto detector to the online-monitor. 
    Data that is written by this plugin should be small (~MB/chunk) 
    to not overload the runs-database.

    This plugin takes 'hitlets_nv' and 'events_nv'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'online_monitor_nv' with info on the hitlets_nv and events_nv
    """
    depends_on = ('hitlets_nv', 'events_nv')
    provides = 'online_monitor_nv'
    data_kind = 'online_monitor_nv'
    rechunk_on_save = False

    # Needed in case we make again an muVETO child.
    ends_with = '_nv'

    __version__ = '0.0.4'

    def infer_dtype(self):
        self.channel_range = self.config['channel_map']['nveto']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.config['events_area_nbins'])

    def compute(self, hitlets_nv, events_nv, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end

        # Count number of hitlets_nv per PMT
        hitlets_channel_count, _ = np.histogram(hitlets_nv['channel'],
                                                bins=self.n_channel,
                                                range=[self.channel_range[0],
                                                       self.channel_range[1] + 1])
        res[f'hitlets{self.ends_with}_per_channel'] = hitlets_channel_count

        # Count number of events_nv with coincidence cut
        res[f'events{self.ends_with}_per_chunk'] = len(events_nv)
        sel = events_nv['n_contributing_pmt'] >= 4
        res[f'events{self.ends_with}_4coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 5
        res[f'events{self.ends_with}_5coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 8
        res[f'events{self.ends_with}_8coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 10
        res[f'events{self.ends_with}_10coinc_per_chunk'] = np.sum(sel)

        # Get histogram of events_nv_area per chunk
        events_area, bins_ = np.histogram(events_nv['area'],
                                          bins=self.config['events_area_nbins'],
                                          range=self.config['events_area_bounds'])
        res[f'events{self.ends_with}_area_per_chunk'] = events_area
        return res


def veto_monitor_dtype(veto_name: str = '_nv',
                       n_pmts: int = 120,
                       n_bins: int = 131) -> list:
    dtype = []
    dtype += strax.time_fields  # because mutable
    dtype += [((f'hitlets{veto_name} per channel', f'hitlets{veto_name}_per_channel'), (np.int64, n_pmts)),
              ((f'events{veto_name}_area per chunk', f'events{veto_name}_area_per_chunk'), np.int64, n_bins),
              ((f'events{veto_name} per chunk', f'events{veto_name}_per_chunk'), np.int64),
              ((f'events{veto_name} 4-coincidence per chunk', f'events{veto_name}_4coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 5-coincidence per chunk', f'events{veto_name}_5coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 8-coincidence per chunk', f'events{veto_name}_8coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 10-coincidence per chunk', f'events{veto_name}_10coinc_per_chunk'), np.int64)
             ]
    return dtype


@export
@strax.takes_config(
    strax.Option(
        'adc_to_pe_mv',
        type=int, default=170.0,
        help='conversion factor from ADC to PE for muon Veto')
)
class OnlineMonitorMV(OnlineMonitorNV):
    __doc__ = OnlineMonitorNV.__doc__.replace('_nv', '_mv').replace('nVeto', 'muVeto')
    depends_on = ('hitlets_mv', 'events_mv')
    provides = 'online_monitor_mv'
    data_kind = 'online_monitor_mv'
    rechunk_on_save = False

    # Needed in case we make again an muVETO child.
    ends_with = '_mv'
    child_plugin = True

    __version__ = '0.0.1'

    def infer_dtype(self):
        self.channel_range = self.config['channel_map']['mv']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.config['events_area_nbins'])

    def compute(self, hitlets_mv, events_mv, start, end):
        events_mv = np.copy(events_mv)
        events_mv['area'] *= 1./self.config['adc_to_pe_mv']
        return super().compute(hitlets_mv, events_mv, start, end)



"""Return corrections from corrections DB
"""
import warnings

import pytz
import numpy as np
from functools import lru_cache
import strax
import utilix
import straxen
import os
from immutabledict import immutabledict


export, __all__ = strax.exporter()

corrections_w_file = ['mlp_model', 'cnn_model', 'gcn_model',
                      's2_xy_map_mlp', 's2_xy_map_cnn', 's2_xy_map_gcn', 's2_xy_map',
                      's1_xyz_map_mlp', 's1_xyz_map_cnn', 's1_xyz_map_gcn',
                      'fdc_map_mlp', 'fdc_map_cnn', 'fdc_map_gcn']

single_value_corrections = ['elife_xenon1t', 'elife', 'baseline_samples_nv',
                            'electron_drift_velocity', 'electron_drift_time_gate']

arrays_corrections = ['hit_thresholds_tpc', 'hit_thresholds_he',
                      'hit_thresholds_nv', 'hit_thresholds_mv']

# needed because we pass these names as strax options which then get paired with the default reconstruction algorithm
# important for apply_cmt_version
posrec_corrections_basenames = ['s1_xyz_map', 'fdc_map', 's2_xy_map']


class CMTVersionError(Exception):
    pass

class CMTnanValueError(Exception):
    pass

@export
class CorrectionsManagementServices():
    """
    A class that returns corrections
    Corrections are set of parameters to be applied in the analysis
    stage to remove detector effects. Information on the strax implementation
    can be found at https://github.com/AxFoundation/strax/blob/master/strax/corrections.py
    """

    def __init__(self, username=None, password=None, mongo_url=None, is_nt=True):
        """
        :param username: corrections DB username
            read the .xenon_config for the users "pymongo_user" has
            readonly permissions to the corrections DB
            the "CMT admin user" has r/w permission to corrections DB
            and read permission to runsDB
        :param password: DB password
        :param is_nt: bool if True we are looking at nT if False we are looking at 1T
        """

        mongo_kwargs = {'url': mongo_url,
                        'user': username,
                        'password': password,
                        'database': 'corrections'}
        corrections_collection = utilix.rundb.xent_collection(**mongo_kwargs)

        # Do not delete the client!
        self.client = corrections_collection.database.client

        # Setup the interface
        self.interface = strax.CorrectionsInterface(
            self.client,
            database_name='corrections')

        self.is_nt = is_nt
        if self.is_nt:
            self.collection = self.client['xenonnt']['runs']
        else:
            self.collection = self.client['run']['runs_new']

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return str(f'{"XENONnT " if self.is_nt else "XENON1T"}'
                   f'-Corrections_Management_Services')

    def get_corrections_config(self, run_id, config_model=None):
        """
        Get context configuration for a given correction
        :param run_id: run id from runDB
        :param config_model: configuration model (tuple type)
        :return: correction value(s)
        """

        if not isinstance(config_model, (tuple, list)) or len(config_model) != 2:
            raise ValueError(f'config_model {config_model} must be a tuple of length 2')
        model_type, version = config_model

        if 'to_pe_model' in model_type:
            return self.get_pmt_gains(run_id, model_type, version)
        elif model_type in single_value_corrections or model_type in arrays_corrections:
            return self._get_correction(run_id, model_type, version)
        elif model_type in corrections_w_file:
            return self.get_config_from_cmt(run_id, model_type, version)
        else:
            raise ValueError(f"{model_type} not found, currently these are "
                             f"available {single_value_corrections}, {arrays_corrections} and "
                             f"{corrections_w_file} ")

    # TODO add option to extract 'when'. Also, the start time might not be the best
    # entry for e.g. for super runs
    # cache results, this would help when looking at the same gains
    @lru_cache(maxsize=None)
    def _get_correction(self, run_id, correction, version):
        """
        Smart logic to get correction from DB
        :param run_id: run id from runDB
        :param correction: correction's name, key word (str type)
        :param version: local version (str type)
        :return: correction value(s)
        """
        when = self.get_start_time(run_id)

        try:
            values = []

            # hack to workaround to group all pmts
            # because every pmt is its own dataframe...of course
            if correction in {'pmt', 'n_veto', 'mu_veto'}:
                # get lists of pmts 
                df_global = self.interface.read('global_xenonnt' if self.is_nt else 'global_xenon1t')
                gains = df_global['global_ONLINE'][0]  # global is where all pmts are grouped
                pmts = list(gains.keys())
                for it_correction in pmts: # loop over all PMTs
                    if correction in it_correction:
                        df = self.interface.read_at(it_correction, when)
                        if df[version].isnull().values.any():
                            raise CMTnanValueError(f"For {it_correction} there are NaN values, this means no correction available "
                                                   f"for {run_id} in version {version}, please check e-logbook for more info ")
 
                        if version in 'ONLINE':
                            df = self.interface.interpolate(df, when, how='fill')
                        else:
                            df = self.interface.interpolate(df, when)
                        values.append(df.loc[df.index == when, version].values[0])
            else:
                df = self.interface.read_at(correction, when)
                if df[version].isnull().values.any():
                    raise CMTnanValueError(f"For {correction} there are NaN values, this means no correction available "
                                           f"for {run_id} in version {version}, please check e-logbook for more info ")
 
                if correction in corrections_w_file or correction in arrays_corrections or version in 'ONLINE':
                    df = self.interface.interpolate(df, when, how='fill')
                else:
                    df = self.interface.interpolate(df, when)
                values.append(df.loc[df.index == when, version].values[0])
            corrections = np.asarray(values)
        except KeyError:
            if "global" in version:
                raise ValueError(f"User is not allowed to pass {version} global version are not allowed")
            raise ValueError(f"Version {version} not found for correction {correction}, please check")
            
        else:
            return corrections

    def get_pmt_gains(self, run_id, model_type, version,
                      cacheable_versions=('ONLINE',),
                      gain_dtype=np.float32):
        """
        Smart logic to return pmt gains to PE values.
        :param run_id: run id from runDB
        :param model_type: to_pe_model (gain model)
        :param version: version
        :param cacheable_versions: versions that are allowed to be
        cached in ./resource_cache
        :param gain_dtype: dtype of the gains to be returned as array
        :return: array of pmt gains to PE values
        """
        to_pe = None
        cache_name = None

        if 'to_pe_model' in model_type:
            # Get the detector name based on the requested model_type
            # This also will be used to the cachable name convention
            # pmt == TPC, n_veto == n_veto's PMT, etc
            detector_names = {'to_pe_model': 'pmt',
                              'to_pe_model_nv': 'n_veto',
                              'to_pe_model_mv': 'mu_veto'}
            target_detector = detector_names[model_type]

            if version in cacheable_versions:
                # Try to load from cache, if it does not exist it will be created below
                cache_name = cacheable_naming(run_id, model_type, version)
                try:
                    to_pe = straxen.get_resource(cache_name, fmt='npy')
                except (ValueError, FileNotFoundError):
                    pass

            if to_pe is None:
                to_pe = self._get_correction(run_id, target_detector, version)

            # be cautious with very early runs, check that not all are None
            if np.isnan(to_pe).any():
                raise ValueError(
                    f"to_pe(PMT gains) values are NaN, no data available "
                    f"for {run_id} in the gain model with version")

        else:
            raise ValueError(f"{model_type} not implemented for to_pe values")

        # Double check the dtype of the gains
        to_pe = np.array(to_pe, dtype=gain_dtype)

        # Double check that all the gains are found, None is not allowed
        # since strax processing does not handle this well. If a PMT is
        # off it's gain should be 0.
        if np.any(np.isnan(to_pe)):
            pmts_affected = np.argwhere(np.isnan(to_pe))[:, 0]
            raise GainsNotFoundError(
                f'Gains returned by CMT are None for PMT_i = {pmts_affected}. '
                f'Cannot proceed with processing. Report to CMT-maintainers.')

        if (cache_name is not None
                and version in cacheable_versions
                and not os.path.exists(cache_name)):
            # This is an array we can save since it's in the cacheable
            # versions but it has not been saved yet. Next time we need
            # it, we can get it from our cache.
            np.save(cache_name, to_pe, allow_pickle=False)
        return to_pe

    def get_config_from_cmt(self, run_id, model_type, version='ONLINE'):
        """
        Smart logic to return NN weights file name to be downloader by
        straxen.MongoDownloader()
        :param run_id: run id from runDB
        :param model_type: model type and neural network type; model_mlp,
        or model_gcn or model_cnn
        :param version: version
        :param return: NN weights file name
        """
        if model_type not in corrections_w_file:
            raise ValueError(f"{model_type} is not stored in CMT "
                             f"please check, these are available {corrections_w_file}")

        file_name = self._get_correction(run_id, model_type, version)

        if not file_name:
            raise ValueError(f"You have the right option but could not find a file"
                             f"Please contact CMT manager and yell at him")
        return file_name

    # TODO change to st.estimate_start_time
    def get_start_time(self, run_id):
        """
        Smart logic to return start time from runsDB
        :param run_id: run id from runDB
        :return: run start time
        """

        if self.is_nt:
            # xenonnt use int
            run_id = int(run_id)

        rundoc = self.collection.find_one(
            {'number' if self.is_nt else 'name': run_id},
            {'start': 1})
        if rundoc is None:
            raise ValueError(f'run_id = {run_id} not found')
        time = rundoc['start']
        return time.replace(tzinfo=pytz.utc)

    def get_local_versions(self, global_version):
        """Returns a dict of local versions for a given global version. Use 'latest' to get newest version"""
        # check that 'global' is in the passed string.

        if global_version == 'latest':
            # CMT appends columns to the global versions dataframe, so taking last one is the latests
            global_version = self.global_versions[-1]

        if 'global' not in global_version:
            warnings.warn("'global' does not appear in the passed global version. Are you sure this right?")
        # CMT generates a global version, global version is just a set of local versions
        # With this we can do pretty easy bookkeping for offline contexts

        cmt_global = self.interface.read('global_xenonnt')
        if global_version not in cmt_global:
            avail_global_versions_string = '\n'.join([f'\t\t{v}' for v in self.global_versions])
            raise ValueError(f"Global version {global_version} not found! "
                             f"Try one of these:\n{avail_global_versions_string}")
        # get local versions from CMT global version
        local_versions = cmt_global[global_version][0]

        # to make returned dictionary more manageable, we prune all the per-PMT corrections
        # first rename to more clear variable
        local_versions['to_pe_model'] = local_versions['pmt_000_gain_xenonnt']
        local_versions['to_pe_model_nv'] = local_versions['n_veto_000_gain_xenonnt']
        local_versions['to_pe_model_mv'] = local_versions['mu_veto_000_gain_xenonnt']

        # drop the per-PMT corrections
        pruned_local_versions = {key: val for key, val in local_versions.items() if "_gain_xenonnt" not in key}
        return pruned_local_versions

    @property
    def global_versions(self):
        return self.interface.read('global_xenonnt').columns.tolist()


def cacheable_naming(*args, fmt='.npy', base='./resource_cache/'):
    """Convert args to consistent naming convention for array to be cached"""
    if not os.path.exists(base):
        try:
            os.mkdir(base)
        except (FileExistsError, PermissionError):
            pass
    for arg in args:
        if not isinstance(arg, str):
            raise TypeError(f'One or more args of {args} are not strings')
    return base + '_'.join(args) + fmt


class GainsNotFoundError(Exception):
    """Fatal error if a None value is returned by the corrections"""


def get_cmt_local_versions(global_version):
    cmt = CorrectionsManagementServices()
    return cmt.get_local_versions(global_version)


@strax.Context.add_method
def apply_cmt_version(context: strax.Context, cmt_global_version: str):
    """Sets all the relevant correction variables
    :param cmt_global_version: A specific CMT global version, or 'latest' to get the newest one
    :returns None
    """
    local_versions = get_cmt_local_versions(cmt_global_version)

    # get the position algorithm we are using
    # I feel like this should be easier...
    posrec_option = 'default_reconstruction_algorithm'
    if posrec_option in context.config:
        posrec_algo = context.config[posrec_option]
    else:
        posrec_algo = context._plugin_class_registry['peak_positions'].takes_config[posrec_option].default

    cmt_options = straxen.get_corrections.get_cmt_options(context)

    # catch here global versions that are not compatible with this straxen version
    # this happens if a new correction was added to CMT that was not used in a fixed version
    # we want this error to occur in order to keep fixed global versions
    cmt_config = dict()
    failed_keys = []
    for option, tup in cmt_options.items():
        try:
            # might need to modify correction name to include position reconstruction algo
            correction_name = tup[0]
            if correction_name in posrec_corrections_basenames:
                correction_name += f"_{posrec_algo}"
            new_tup = (tup[0], local_versions[correction_name], tup[2])
        except KeyError:
            failed_keys.append(option)
            continue
        cmt_config[option] = new_tup
    if len(failed_keys):
        failed_keys = ', '.join(failed_keys)
        raise CMTVersionError(f"CMT version {cmt_global_version} is not compatible with this straxen version! "
                              f"CMT {cmt_global_version} is missing these corrections: {failed_keys}")

    context.set_config(cmt_config)



import bokeh.plotting as bklt
from bokeh.embed import file_html
from bokeh.resources import CDN
import numpy as np
import strax

export, __all__ = strax.exporter()


@export
def bokeh_to_wiki(fig, outputfile=None):
    """
    Function which converts bokeh HTML code to a wiki readable code.

    :param fig: Figure to be conerted
    :param outputfile: String of absolute file path. If specified output
        is writen to the file. Else output is print to the notebook and
        can be simply copied into the wiki.
    """
    # convert plot to wiki format:
    html = file_html(fig, CDN)
    html = '\n'.join((['<html>'] + html.split('\n')[6:]))

    if outputfile:
        with open(outputfile, mode='w') as file:
            file.write(html)
    else:
        print(html)


def get_peaks_source(peaks, relative_start=0, time_scaler=1, keep_amplitude_per_sample=True):
    """
    Computes bokeh.plotting.ColumnDataSource for given peaks.

    :param peaks: Peaks to be plotted.
    :param relative_start: t0 from which on the peaks should be plotted.
    :param time_scaler: Factor to rescale the time from ns to other scale.
        E.g. =1000 scales to µs.
    :param keep_amplitude_per_sample: Boolean if true amplitude of the
        plotted peaks is in pe/sample. False pe/ns.
    :return: bokeh.plotting.ColumnDataSource instance which can be used
        to plot peaks.
    """
    if not (np.all(peaks['type'] == peaks[0]['type'])):
        raise ValueError('All peaks must be of the same type (S1, S2 or Unknown)!')

    x_p = []
    y_p = []
    for ind, p in enumerate(peaks):
        x, y = _patches_x_y(p, keep_amplitude_per_sample=keep_amplitude_per_sample)
        x -= relative_start  # relative to first peak
        x = x / time_scaler
        x_p.append(x)
        y_p.append(y)

    if peaks[0]['type'] == 2:
        scaler = 10**-3
    else:
        scaler = 1

    source = bklt.ColumnDataSource(data={'xs': x_p,  # Coordinates for peak patches
                                         'ys': y_p,
                                         'x': peaks['x'],  # XY-Pos in PMZ Hitpattern
                                         'y': peaks['y'],
                                         'dt': peaks['dt'],
                                         'time': peaks['time'],
                                         'center_time': peaks['center_time'],
                                         'endtime': strax.endtime(peaks),
                                         'width_50': peaks['range_50p_area'] * scaler,
                                         'width_90': peaks['range_90p_area'] * scaler,
                                         'rise': peaks['rise_time'] * scaler,
                                         'rel_center_time': (peaks['center_time'] - peaks['time']) * scaler,
                                         'area': peaks['area'],
                                         'aft': peaks['area_fraction_top'],
                                         'nhits': peaks['n_hits'],
                                         }
                                   )
    return source


def _patches_x_y(peak, keep_amplitude_per_sample=False):
    """
    Creates x,y coordinates needed to draw peaks via
    bokeh.models.patches.

    :param peak: Peak for which we need the x/y samples
    :param keep_amplitude_per_sample: If y-data should be in units of "per sample"
        or "per ns".

    :returns: Tuple of x, y
    """
    if keep_amplitude_per_sample:
        dt_a = 1
    else:
        dt_a = peak['dt']

    x = np.arange(peak["length"])
    xx = np.zeros(2 * len(x), dtype=x.dtype)
    mx = 0.5 * (x[1::] + x[:-1])
    xx[1:-1:2] = mx
    xx[2::2] = mx
    xx[0] = 1.5 * x[0] - 0.5 * x[1]
    xx[-1] = 1.5 * x[-1] - 0.5 * x[-2]
    xx = np.array(xx) * peak['dt'] + peak['time']
    y = peak['data'][:peak['length']]
    yy = np.zeros(2 * len(x))
    yy[0::2] = y
    yy[1::2] = y
    yy = np.array(yy) / dt_a

    # baseline since we'll fill underneath
    xx = np.concatenate([[xx[0]], xx, [xx[-1]]])
    yy = np.concatenate([[0], yy, [0]])
    return xx, yy


def peak_tool_tip(peak_type):
    """
    Default mouseover tooltip for peaks.

    :param peak_type: If 2, all time variables are in µs else in ns.
    :return: dictionary of tooltips. Can be converted to a list for
        bokeh.models.HoverTool.
    """

    # Add static time parameters:
    tool_tip = {"time_static": ("time [ns]", "@time"),
                "center_time": ("center_time [ns]", "@center_time"),
                "endtime": ("endtime [ns]", "@endtime"),
                }

    # Now ns/µs parameters for S1 and S2
    tool_tip['dt'] = ("dt [ns/sample]", "@dt")
    tool_tip['time_dynamic'] = ("time [ns]", "$x")
    tool_tip['rel_center_time'] = ('center time [ns]', '@rel_center_time')
    tool_tip['range_50p_width'] = ('50% width [ns]', '@width_50')
    tool_tip['range_90p_width'] = ('90% width [ns]', '@width_90')
    tool_tip['rise_time'] = ('rise time [ns]', '@rise')

    # Add non-time parameters (results in an ordered tooltip)
    tool_tip['amplitude'] = ("Amplitude [pe/ns]", "$y")
    tool_tip["area"] = ('area [pe]', '@area')
    tool_tip["aft"] = ('AFT', '@aft')
    tool_tip["nhits"] = ('nhits', '@nhits')

    if peak_type == 2:
        for k, i in tool_tip.items():
            if k not in ["time_static", "center_time", "endtime"]:
                tool_tip[k] = (i[0].replace('[ns]', '[µs]'), i[1])

    return tool_tip


def default_fig(width=400, height=400, title='', **kwargs):
    """
    Helper function which returns a bokeh.plotting.figure instance
    with sizing_mode set to 'scale_both' and an aspect ratio set
    according to the specified width and height.

    :param width: Plot width in pixels
    :param height: PLot height in pixels.
    :param title: Title of the plot.

    Also allows for additional kwargs accepted by bokeh.plotting.

    :returns: bokeh.plotting.figure  instance.
    """
    fig = bklt.figure(plot_width=width,
                      plot_height=height,
                      sizing_mode='scale_both',
                      aspect_ratio=width / height,
                      title=title,
                      tools="pan,box_zoom,reset,save",
                      **kwargs
                      )
    return fig



import numpy as np
import strax
import numba
export, __all__ = strax.exporter()

@export
@strax.takes_config(
    strax.Option('pre_s2_area_threshold', default=1000, infer_type=False,
                 help='Only take S2s large than this into account when calculating EventShadow.'),
    strax.Option('time_window_backward', default=int(3e9), infer_type=False,
                 help='Search for S2s causing shadow in this time window'))
class EventShadow(strax.Plugin):
    """
    This plugin can find and calculate the previous S2 shadow at event level,
    with time window backward and previous S2 area as options.
    It also gives the area and position infomation of these previous S2s.
    """
    __version__ = '0.0.6'
    depends_on = ('event_basics','peak_basics','peak_positions')
    provides = 'event_shadow'
    save_when = strax.SaveWhen.EXPLICIT
    

    def infer_dtype(self):
        dtype = [
                ('pre_s2_area', np.float32,'previous s2 area [PE]'),
                 ('shadow_dt', np.int64,'time diffrence to the previous s2 [ns]'),
                 ('shadow', np.float32,'previous s2 shadow [PE/ns]'),
                ('pre_s2_x', np.float32,'x of previous s2 peak causing shadow [cm]'),
                 ('pre_s2_y', np.float32,'y of previous s2 peak causing shadow [cm]'),
                 ('shadow_distance', np.float32,'distance to the previous s2 peak causing the max shadow [cm]')           
                ]
        dtype += strax.time_fields
        return dtype
    
    def compute(self, events, peaks):
        roi_dt = np.dtype([(('back in time', 'time'), int),
                                         (('till it begin','endtime'), int)])
        roi = np.zeros(len(events), dtype=roi_dt)   
        n_seconds = self.config['time_window_backward']
        roi['time'] = events['time'] - n_seconds
        roi['endtime'] = events['time']
        mask_s2 = peaks['type'] == 2
        mask_s2  &= peaks['area'] > self.config['pre_s2_area_threshold']   
        split_peaks = strax.split_touching_windows(peaks[mask_s2], roi)
        res = np.zeros(len(events), self.dtype)       
        compute_shadow(events, split_peaks, res)
        res['shadow_distance'] = ((res['pre_s2_x'] - events['s2_x'])**2+(res['pre_s2_y'] - events['s2_y'])**2)**0.5
        res['time'] = events['time']
        res['endtime'] = strax.endtime(events)
        
        return res
    
def compute_shadow(events, split_peaks, res):
    if len(res):
        return _compute_shadow(events, split_peaks, res)


@numba.njit(cache=True)
def _compute_shadow(events, split_peaks, res):
    for event_i, event_a in enumerate(events):
        new_shadow = 0
        for peak_i, peak_a in enumerate(split_peaks[event_i]):
            new_shadow = peak_a['area']/(event_a['s2_center_time']-peak_a['center_time'])
            if new_shadow > res['shadow'][event_i]:
                res['pre_s2_area'][event_i] = peak_a['area']
                res['shadow_dt'][event_i] = event_a['s2_center_time']-peak_a['center_time']
                res['pre_s2_x'][event_i] = peak_a['x']
                res['pre_s2_y'][event_i] = peak_a['y']
                res['shadow'][event_i] = new_shadow     



import strax
import straxen
import numba
import numpy as np

export, __all__ = strax.exporter()

__all__ += ['T_NO_VETO_FOUND']
# Runs are usually 1 hour long, if veto is that far we don't really care
T_NO_VETO_FOUND = int(3.6e+12)


# More info about the acquisition monitor can be found here:
# https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenon1t:alexelykov:acquisition_monitor


@export
@strax.takes_config(strax.Option('hit_min_amplitude_aqmon', default=50, track=True, infer_type=False,
                                 help='Minimum hit threshold in ADC*counts above baseline'),
                    strax.Option('baseline_samples_aqmon', default=10, track=True, infer_type=False,
                                 help='Number of samples to use at the start of the pulse to determine the baseline'))
class AqmonHits(strax.Plugin):
    """ Find hits in acquisition monitor data. These hits could be 
        then used by other plugins for deadtime calculations, 
        GPS SYNC analysis, etc.
    """
    __version__ = '0.0.6'

    depends_on = ('raw_records_aqmon')
    provides = ('aqmon_hits')
    data_kind = ('aqmon_hits')

    dtype = strax.hit_dtype

    save_when = strax.SaveWhen.TARGET

    def compute(self, raw_records_aqmon):
        rec = strax.raw_to_records(raw_records_aqmon)
        strax.sort_by_time(rec)
        strax.zero_out_of_bounds(rec)
        strax.baseline(rec, baseline_samples=self.config['baseline_samples_aqmon'], flip=True)
        aqmon_hits = strax.find_hits(rec, min_amplitude=self.config['hit_min_amplitude_aqmon'])
        return aqmon_hits


# ### Veto hardware ###:
# V1495 busy veto module:
# Generates a 25 ns NIM pulse whenever a veto begins and a 25 ns NIM signal when it ends.
# A new start signal can occur only after the previous busy instance ended.
# 1ms (1e6 ns) - minimum busy veto length, or until the board clears its memory

# DDC10 High Energy Veto:
# 10ms (1e7 ns) - fixed HE veto length in XENON1T DDC10,
# in XENONnT it will be calibrated based on the length of large S2 SE tails
# The start/stop signals for the HEV are generated by the V1495 board


@export
@strax.takes_config(
    strax.Option('max_veto_window', default=int(5e8), track=True, type=int,
                 help='Maximum separation between veto stop and start pulses [ns]'),
)
class VetoIntervals(strax.OverlapWindowPlugin):
    """ Find pairs of veto start and veto stop signals and the veto duration between them
    busy_*  <= V1495 busy veto for tpc channels
    he_*    <= V1495 busy veto for high energy tpc channels
    hev_*   <= DDC10 hardware high energy veto
    """

    __version__ = '0.1.6'
    depends_on = ('aqmon_hits')
    provides = ('veto_intervals')
    data_kind = ('veto_intervals')

    def infer_dtype(self):
        dtype = [(('veto interval [ns]', 'veto_interval'), np.int64),
                 (('veto signal type', 'veto_type'), np.str_('U9'))]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.veto_names = ['busy_', 'he_', 'hev_']
        self.channel_map = {name: ch + straxen.n_hard_aqmon_start for ch, name in
                            enumerate(['sum_wf', 'm_veto_sync',
                                       'hev_stop', 'hev_start',
                                       'he_stop', 'he_start',
                                       'busy_stop', 'busy_start'])}

    def get_window_size(self):
        # Give a very wide window
        return int(self.config['max_veto_window'] * 100)

    def compute(self, aqmon_hits, start, end):
        hits = aqmon_hits
        result = np.zeros(len(aqmon_hits) * len(self.veto_names), self.dtype)
        vetos_seen = 0

        for name in self.veto_names:
            veto_hits_start = channel_select_(hits, self.channel_map[name + 'start'])
            veto_hits_stop = channel_select_(hits, self.channel_map[name + 'stop'])

            # Here we rely on the fact that for each start, there is a single stop that
            # follows it in time. If this is not true, our hardware does not work.
            if len(veto_hits_start):
                for t, time in enumerate(veto_hits_start['time']):
                    # Find the time of stop_j that is closest to time of start_i
                    inx = np.searchsorted(veto_hits_stop['time'], time, side='right')

                    if inx == len(veto_hits_stop['time']):
                        continue
                    else:
                        result['veto_interval'][vetos_seen] = veto_hits_stop['time'][inx] - time
                        result["time"][vetos_seen] = time
                        assert time < veto_hits_stop['time'][inx]
                        result["endtime"][vetos_seen] = veto_hits_stop['time'][inx]
                        result["veto_type"][vetos_seen] = name + 'veto'
                        vetos_seen += 1
        result = result[:vetos_seen]
        result['time'] = np.clip(result['time'], start, end)
        result['endtime'] = np.clip(strax.endtime(result), 0, end)
        sort = np.argsort(result['time'])
        result = result[sort]
        return result


@numba.njit
def channel_select_(rr, ch):
    """Return data from start/stop veto channel in the acquisition monitor (AM)"""
    return rr[rr['channel'] == ch]


@export
@strax.takes_config(
    strax.Option('veto_proximity_window', default=int(5e8), type=int, track=True,
                 help='Maximum separation between veto stop and start pulses [ns]'))
class VetoProximity(strax.OverlapWindowPlugin):
    """
    Find the closest next/previous veto start and end to each event center.

    previous_x: Time in ns between the time center of an event and the previous x
    next_x: Time in ns between the time center of an event and the next x
    This also considers any x inside the event. x could be either:
        - busy_x: busy on/off signal
        - he_x:   high energy channels busy on/off signal
        - hev_x:  high energy veto on/off signal
    """

    __version__ = '0.1.2'
    depends_on = ('event_basics', 'aqmon_hits')
    provides = ('veto_proximity')
    data_kind = ('events')
    save_when = strax.SaveWhen.TARGET
    veto_names = ['busy', 'he', 'hev']

    def infer_dtype(self):
        dtype = []
        for n in self.veto_names:
            dtype += [
                ((f'Time to previous {n} veto start [ns]', f'previous_{n}_on'), np.int64),
                ((f'Time to previous {n} veto end [ns]', f'previous_{n}_off'), np.int64),
                ((f'Time to next {n} veto start [ns]', f'next_{n}_on'), np.int64),
                ((f'Time to next {n} veto end [ns]', f'next_{n}_off'), np.int64),
                ]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.channel_map = {name: ch + straxen.n_hard_aqmon_start for ch, name in
                            enumerate(['sum_wf', 'm_veto_sync',
                                       'hev_stop', 'hev_start',
                                       'he_stop', 'he_start',
                                       'busy_stop', 'busy_start'])}
        self.states = ['on', 'off']

    def get_window_size(self):
        return int(self.config['veto_proximity_window'] * 100)

    def compute(self, events, aqmon_hits):
        result = np.zeros(len(events), self.dtype)
        t_event_centers = (events['time'] + events['endtime']) // 2

        for name in self.veto_names:
            # For each state find the next and previous veto
            for state in self.states:
                prev = f'previous_{name}_{state}'
                nxt = f'next_{name}_{state}'

                if state == 'on':
                    aqmon_chan = self.channel_map[f'{name}_start']
                else:
                    aqmon_chan = self.channel_map[f'{name}_stop']
                veto_start_time_selection = aqmon_hits[aqmon_hits['channel'] == aqmon_chan]['time']

                inx = 0
                for event_i, event_center in enumerate(t_event_centers):
                    if len(veto_start_time_selection):
                        inx = np.searchsorted(veto_start_time_selection, event_center, side='right')

                    # Time to previous veto on/off
                    # Just using a huge value that will not fit in any potential
                    # DAQVetoCut range
                    if inx == 0:
                        previous_veto = T_NO_VETO_FOUND
                    else:
                        previous_veto = event_center - veto_start_time_selection[inx - 1]
                    # Time to next veto on/off
                    if inx == len(veto_start_time_selection):
                        next_veto = T_NO_VETO_FOUND
                    else:
                        next_veto = veto_start_time_selection[inx] - event_center

                    result[event_i][prev] = previous_veto
                    result[event_i][nxt] = next_veto

        # Add the events time and endtime to the final result
        result['time'] = events['time']
        result['endtime'] = events['endtime']
        return result



import numpy as np
import strax
export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option('n_tpc_pmts', type=int,
                 help='Number of TPC PMTs'),
)
class EventAreaPerChannel(strax.LoopPlugin):
    """
    Simple plugin that provides area per channel for main 
    and alternative S1/S2 in the event. 
    """
    depends_on = ('event_basics', 'peaks')
    provides = "event_area_per_channel"
    __version__ = '0.0.0'

    compressor = 'zstd'
    save_when = strax.SaveWhen.EXPLICIT

    def infer_dtype(self):
        dtype = [(("Area per channel for main S2", "s2_area_per_channel"), 
                  np.float32, (self.config['n_tpc_pmts'],)),
                 (("Area per channel for alternative S2", "alt_s2_area_per_channel"),
                  np.float32, (self.config['n_tpc_pmts'],)),
                 (("Area per channel for main S1", "s1_area_per_channel"), 
                  np.float32, (self.config['n_tpc_pmts'],)),
                 (("Area per channel for alternative S1", "alt_s1_area_per_channel"), 
                  np.float32, (self.config['n_tpc_pmts'],)),
                ]
        dtype += strax.time_fields
        return dtype
    
    def compute_loop(self, event, peaks):
        result = dict()
        result['time'] = event['time']
        result['endtime'] = strax.endtime(event)
        
        for type_ in ['s1', 's2', 'alt_s1', 'alt_s2']:
            type_index = event[f'{type_}_index']
            if type_index != -1:
                type_ara_per_channel = peaks['area_per_channel'][type_index]
                result[f'{type_}_area_per_channel'] = type_ara_per_channel
        return result



import strax
import straxen
from straxen.analyses.holoviews_waveform_display import seconds_from, plot_record_polygons

import panel as pn
from bokeh.models import HoverTool

import numpy as np
import pandas as pd

export, __all__ = strax.exporter()


@export
class nVETOEventDisplay:
    def __init__(self,
                 events=None,
                 hitlets=None,
                 run_id=0,
                 channel_range=(2000, 2119),
                 pmt_map='nveto_pmt_position.csv',
                 plot_extension='bokeh'):
        """
        Class to plot an interactive nveto display.

        :param events: Events which should be plot. Can also be none
            in case the hitlet matrix and/or pattern map should be
            plotted separately.
        :param hitlets: Same as events, but hitlets_nv.
        :param run_id: Run_id which should be displayed in the title.
        :param channel_range: Channel range of the detector.
        :param pmt_map: PMT map which is loaded via straxen.get_resource.
            The map has to contain the channel number, and xyz
            coordinates.
        :param plot_extension: Extension which should be used for
            rendering can be either bokeh or matpltolib. Default is
            bokeh to support dynamic plots.
        """
        self.import_holoviews()
        self.hv.extension(plot_extension)
        self.df_event_time = None
        self.df_event_properties = None
        self.hitlets = hitlets
        self.channel_range = channel_range
        self.run_id = run_id

        # Load PMT data:
        if isinstance(pmt_map, str):
            self.pmt_positions = straxen.get_resource(pmt_map,
                                                      fmt='csv')
        elif isinstance(pmt_map, np.ndarray):
            self.pmt_positions = pmt_map
        else:
            raise ValueError('pmt_map not understood, has either to be '
                             f'a string or a numpy array, got "{pmt_map}".')

        if events is not None:
            self.event_df = straxen.convert_array_to_df(events)
        else:
            self.event_df = None

        if events is not None and hitlets is not None:
            self.hitlets_per_event = strax.split_by_containment(hitlets,
                                                                events)

    def import_holoviews(self):
        import holoviews as hv
        self.hv = hv

    def plot_event_display(self):
        """
        Creates an interactive event display for the neutron veto.

        :returns: panel.Column hosting the plots and panels.
        """
        # First we have to define the python callbacks:
        def matrix_callback(value):
            """
            Callback for the dynamic hitlet matrix. Changes polygons when a new
            event is selected.
            """
            self.hitlet_points = self.hitlets_to_hv_points(self.hitlets_per_event[value],
                                                           t_ref=self.event_df.loc[value, 'time']
                                                           )

            # Create the hitlet matrix and time stream:
            self.hitlet_matrix = self.plot_hitlet_matrix(hitlets=None,
                                                         _hitlet_points=self.hitlet_points)
            return self.hitlet_matrix

        def pattern_callback(value, x_range):
            """
            Call back for the dynamic PMT pattern map. Depends on the selcted
            event as well as the selected x_range in the hitlet_matrix.
            """
            # Get hitlet points and select only points within x_range:
            hit = self.hitlet_points.data
            if not x_range:
                time = hit['time'].values
                length = hit['length'].values
                dt = hit['dt'].values
                x_range = [min(time), max(time + (length * dt))]

            m = (hit['time'] >= x_range[0]) & (hit['time'] < x_range[1])

            hitlets_in_time = hit[m]
            new_points = self.hv.Points(hitlets_in_time)

            # Plot pmt pattern:
            pmts = self.plot_nveto(hitlets=None,
                                   _hitlet_points=new_points,
                                   pmt_size=8,
                                   pmt_distance=0.5)
            angle = self._plot_reconstructed_position(value)
            return angle * pmts

        self._make_sliders_and_tables(self.event_df)
        index = self.evt_sel_slid.value
        self.hitlet_points = self.hitlets_to_hv_points(self.hitlets_per_event[index],
                                                       t_ref=self.event_df.loc[index, 'time'])

        dmap_hitlet_matrix = self.hv.DynamicMap(
            matrix_callback,
            streams=[self.evt_sel_slid.param.value]).opts(framewise=True)

        time_stream = self.hv.streams.RangeX(source=dmap_hitlet_matrix)

        dmap_pmts = self.hv.DynamicMap(
            pattern_callback,
            streams=[self.evt_sel_slid.param.value, time_stream])

        slider_column = pn.Column(self.evt_sel_slid,
                                  self.evt_sel_slid.controls(['value']),
                                  self.time_table)

        event_display = pn.Column(self.title_panel,
                                  pn.Row(slider_column, dmap_pmts, width_policy='max'),
                                  dmap_hitlet_matrix,
                                  self.prop_table,
                                  self.pos_table,
                                  width_policy='max')
        return event_display

    def plot_hitlet_matrix(self, hitlets, _hitlet_points=None):
        """
        Function which plots the hitlet matrix for the specified hitlets.
        The hitlet matrix is something equivalent to the record matrix
        for the TPC.

        :param hitlets: Hitlets to be plotted if called directly.
        :param _hitlet_points: holoviews.Points created by the event
            display. Only internal use.
        :returns: hv.Polygons plot.
        """
        if not _hitlet_points:
            _hitlet_points = self.hitlets_to_hv_points(hitlets, )

        hitlet_matrix = self._plot_base_matrix(_hitlet_points).opts(title='Hitlet Matrix',
                                                                    xlabel='Time [ns]',
                                                                    ylabel='PMT channel',
                                                                    ylim=(1995, 2125),
                                                                    color='area',
                                                                    clabel='Area [pe]',
                                                                    cmap='viridis',
                                                                    colorbar=True
                                                                    )
        return hitlet_matrix

    def _plot_base_matrix(self, hv_point_data):
        """
        Base function to plot record or hitlet matrix.
        """
        matrix_plot = self.hv.Segments(hv_point_data.data,
                                  kdims=['time', 'channel', 'endtime', 'channel']
                                  ).opts(tools=['hover'],
                                         aspect=4,
                                         responsive='width',
                                         )
        return matrix_plot

    def plot_nveto(self,
                   hitlets,
                   pmt_size=8,
                   pmt_distance=0.5,
                   _hitlet_points=None, ):
        """
        Plots the nveto pmt pattern map for the specified hitlets.
        Expects hitlets to be sorted in time.

        :param hitlets: Hitlets to be plotted if called directly.
        :param pmt_size: Base size of a PMT for 1 pe.
        :param pmt_distance: Scaling parameter for the z -> xy
            projection.
        :param _hitlet_points: holoviews.Points created by the event
            display. Only internal use.

        :returns: stacked hv.Points plot.
        """
        if not _hitlet_points:
            _hitlet_points = self.hitlets_to_hv_points(hitlets, )

        pmts = self._plot_nveto(_hitlet_points.data,
                                pmt_size=pmt_size,
                                pmt_distance=pmt_distance)

        # Plot channels which have not seen anything:
        pmts_data = pmts.data
        m = pmts_data.loc[:, 'area'].values == 0
        blanko_pmts = np.zeros(np.sum(m), dtype=strax.hitlet_dtype())
        blanko_pmts['area'] = 1
        blanko_pmts['channel'] = pmts_data.loc[m, 'channel']
        blanko = self._plot_nveto(blanko_pmts,
                                  pmt_size=pmt_size,
                                  pmt_distance=pmt_distance)

        blanko = blanko.opts(fill_color='white',
                             tools=[HoverTool(tooltips=[('PMT', '@channel')])],
                             color=None)

        detector_layout = (plot_tpc_circle(straxen.cryostat_outer_radius)
                           * plot_diffuser_balls_nv()
                           * plot_nveto_reflector())

        return detector_layout * blanko * pmts

    def _plot_nveto(self,
                    pmts,
                    pmt_size,
                    pmt_distance):
        if isinstance(pmts, pd.DataFrame):
            pmts = pmts.to_records(index=False)
        hv = self.hv
        # Get hitlet_points data and extract PMT data
        pmt_data = self._create_nveto_pmts(pmts,
                                           pmt_size,
                                           pmt_distance)

        pmts = hv.Points(data=pd.DataFrame(pmt_data),
                         kdims=[hv.Dimension('pmt_x',
                                             label='x [cm]'),
                                hv.Dimension('pmt_y',
                                             label='y [cm]')],
                         vdims=['time', 'size', 'area', 'channel']
                         ).opts(size='size',
                                tools=[self._create_pmt_pattern_hover()],
                                color='time',
                                colorbar=True,
                                cmap='viridis',
                                clabel='Time [ns]',
                                line_color='black',
                                alpha=0.8,
                                width=400,
                                title='nVETO Top View')

        return pmts

    def _create_nveto_pmts(self,
                           hitlets,
                           pmt_size,
                           pmt_distance,
                           max_area_scale=10, ):
        """
        Function which creates data for the nVTEO PMT points of the
        pattern plot.
        """
        # Get PMT xy-position based on projection function:
        pmt_data = self._convert_channel_to_xy(pmt_distance=pmt_distance)

        for h in hitlets:
            ch = h['channel'] - self.channel_range[0]
            pmt_data[ch]['area'] += h['area']

            if not pmt_data[ch]['time']:
                # Get first arrival time for channel:
                pmt_data[ch]['time'] = h['time']

        pmt_data['size'] = pmt_data['area'] * pmt_size
        pmt_data['size'] = np.clip(pmt_data['size'], 0, max_area_scale * pmt_size)

        return pmt_data

    def _convert_channel_to_xy(self, pmt_distance=0.5):
        """
        Projects PMT channel from x/y/z onto xy.
        """
        pmt_data = np.zeros(120, dtype=[('pmt_x', np.float64),
                                        ('pmt_y', np.float64),
                                        ('channel', np.int32),
                                        ('column', np.int32),
                                        ('area', np.float64),
                                        ('size', np.float64),
                                        ('time', np.float64), ])
        pmt_data['pmt_x'] = self.pmt_positions['x']
        pmt_data['pmt_y'] = self.pmt_positions['y']
        pmt_data['channel'] = self.pmt_positions['channel']
        pmt_data['column'] = np.repeat(np.arange(20), 6)

        z = self.pmt_positions['z']
        z_min = self.pmt_positions['z'].min()
        z_distance = self.pmt_positions['z'].max() - z_min
        scale = (z - z_min) / z_distance * pmt_distance

        # Only correct x or y depending on column position:
        m_x = np.isin(pmt_data['column'], (4, 5, 6, 7, 8, 14, 15, 16, 17, 18))
        m_y = np.isin(pmt_data['column'], (19, 0, 1, 2, 3, 9, 10, 11, 12, 13))

        pmt_data['pmt_x'][m_x] = pmt_data['pmt_x'][m_x] * (1 + scale[m_x])
        pmt_data['pmt_y'][m_y] = pmt_data['pmt_y'][m_y] * (1 + scale[m_y])

        return pmt_data

    @staticmethod
    def _create_pmt_pattern_hover():
        tooltips = [('rel. time [ns]', '@time'),
                    ('integrated area [pe]', '@area'),
                    ('PMT', '@channel'),
                    ]
        return HoverTool(tooltips=tooltips)

    @staticmethod
    def hitlets_to_hv_points(hitlets,
                             t_ref=None, ):
        """
        Function which converts hitlets into hv.Points used in the
        different plots. Computes hitlet times as relative times with
        respect to the first hitlet if t_ref is not set.
        """
        import holoviews as hv
        if not len(hitlets):
            raise ValueError('Expected at least a single hitlet.')

        if isinstance(hitlets, np.ndarray):
            hitlets = pd.DataFrame(hitlets)

        # Set relative times:
        if t_ref is None:
            t_ref = min(hitlets['time'])

        time = seconds_from(hitlets['time'],
                            t_ref,
                            unit_conversion=1)
        hitlets['time'] = time
        hitlets['endtime'] = strax.endtime(hitlets.to_records())

        hitlet_points = hv.Points(hitlets)

        return hitlet_points

    def _plot_reconstructed_position(self, index):
        """
        Function which plots the nVETO event position according to its
        azimuthal angle.

        :param index: Which event to plot.
        """

        x = (0, np.real(np.exp(self.event_df.loc[index, 'angle'] * 1j)) * 400)
        y = (0, np.imag(np.exp(self.event_df.loc[index, 'angle'] * 1j)) * 400)
        angle = self.hv.Curve((x, y)).opts(
            color='orange',
            line_dash='dashed',
            xlim=(-350, 350),
            ylim=(-350, 350))
        return angle

    def _make_sliders_and_tables(self, df):
        """
        Function which creates interactive sliders and tables for the
        neutron-veto event display
        """
        if not len(df):
            raise ValueError('DataFrame must be at least one entry long.')

        self.evt_sel_slid = pn.widgets.IntSlider(value=0,
                                                 start=0,
                                                 end=len(df))
        self._make_tables(df)

        # Define callbacks for tables:
        self.evt_sel_slid.param.watch(lambda event: table_callback(self.time_table,
                                                                   self.df_event_time,
                                                                   event, True),
                                      'value')
        self.evt_sel_slid.param.watch(lambda event: table_callback(self.prop_table,
                                                                   self.df_event_properties,
                                                                   event),
                                      'value')
        self.evt_sel_slid.param.watch(lambda event: table_callback(self.pos_table,
                                                                   self.df_event_position,
                                                                   event),
                                      'value')

        # Now make title and also define callback:
        title = self._make_title(self.evt_sel_slid.value)
        self.title_panel = pn.panel(title, sizing_mode='scale_width')

        def title_callback(event):
            self.title_panel.object = self._make_title(event.new)

        self.evt_sel_slid.param.watch(title_callback, 'value')

    def _make_tables(self, df):
        """
        Divides event data into a time, properties and position table.
        """
        # Time table:
        time_keys = ['time', 'endtime', 'event_number_nv']
        self.df_event_time = df.loc[:, time_keys]

        # Properties tables:
        pos_keys = ['angle', 'pos_x', 'pos_x_spread', 'pos_y',
                    'pos_y_spread', 'pos_z', 'pos_z_spread']
        self.df_event_position = df.loc[:, pos_keys]

        keys = df.columns.values
        keys = [k for k in keys if k not in time_keys + pos_keys]
        self.df_event_properties = df.loc[:, keys]

        # Table panels:
        index = self.evt_sel_slid.value
        self.time_table = pn.panel(self.df_event_time.loc[index],
                                   )
        self.pos_table = pn.panel(self.df_event_position.loc[index:index, :],
                                  sizing_mode='scale_width')

        self.prop_table = pn.panel(self.df_event_properties.loc[index:index, :],
                                   sizing_mode='scale_width')

    def _make_title(self, ind):
        """
        Function which creates title test in markdown format.
        """
        start = self.df_event_time.loc[ind, 'time']
        date = np.datetime_as_string(start.astype('<M8[ns]'), unit='s')
        start_ns = start - (start // 10**9) * 10**9
        end = self.df_event_time.loc[ind, 'endtime']
        end_ns = end - start + start_ns
        return ''.join((f'##Event {ind} from run {self.run_id}\n',
                        f'##Recorded at ({date[:10]} {date[10:]}) UTC ',
                        f'{start_ns} ns - {end_ns} ns'))


def plot_tpc_circle(radius):
    """
    Plots TPC as a black circle

    :param radius: Radius in cm.
    """
    import holoviews as hv
    x = radius * np.cos(np.arange(-np.pi, np.pi + 0.1, 0.01))
    y = radius * np.sin(np.arange(-np.pi, np.pi + 0.1, 0.01))
    return hv.Curve((x, y)).opts(color='k')


def plot_diffuser_balls_nv():
    """
    Computes position of nveto diffuser balls.

    :return: hv.Points with hover tool.
    """
    import holoviews as hv
    cryostat_r = straxen.cryostat_outer_radius
    depth_stiffening_ring = 8
    r = cryostat_r + depth_stiffening_ring
    angles = np.array([-8, 90, -90, 180])
    x = r * np.cos(angles / 180 * np.pi)
    y = r * np.sin(angles / 180 * np.pi)
    db_ids = (18, 17, 11, 15)
    data = pd.DataFrame(np.array([x, y, db_ids]).T, columns=('x', 'y', 'id'))
    return hv.Points(data=data).opts(size=8, tools=['hover'])


def plot_nveto_reflector():
    """
    Function which plots the lateral reflector panels of the neutron
    veto. Coordinates are based on the MC coordinates of the octagon
    model. The coordinates can be found in this note:
    id=xenon:xenonnt:mc:notes:nveto-geometry#lateral_reflectors
    """
    import holoviews as hv
    xy_center_angle = np.array([(0, -1875.25, -90),
                                (1442.28, -1442.28, -45),
                                (1875.25, 0, 0),
                                (1442.28, 1442.28, 45),
                                (0, 1875.25, 90),
                                (-1442.28, 1442.28, 135),
                                (-1875.25, 0, 180),
                                (-1441.75, -1442.28, -135),
                                (0, -1875.25, -90)], dtype=[('x', np.float64),
                                                            ('y', np.float64),
                                                            ('phi', np.float64), ]
                               )
    xy_pos = _compute_lateral_reflector_xy_edges(xy_center_angle,
                                                 long_side_length=2018,
                                                 short_side_length=1224)
    xy_pos = pd.DataFrame(xy_pos, columns=('x', 'y'))

    return hv.Curve(data=xy_pos).opts(color='k')


def _compute_lateral_reflector_xy_edges(xy_center_angle,
                                        long_side_length=2018,
                                        short_side_length=1224):
    """
    Function which computes the position of the lateral reflector
    panels. Input in mm return in cm.

    :param xy_center_angle: Center xy coordinate and angle of each
        panel.
    :param long_side_length: Full length of the long panels in mm.
    :param short_side_length: Same
    """
    res = np.zeros(len(xy_center_angle), dtype=[('x', np.float64),
                                                ('y', np.float64)])
    for ind, xy in enumerate(xy_center_angle):
        if ind % 2:
            length = short_side_length / 2
        else:
            length = long_side_length / 2
        xc, yc, phi = xy[['x', 'y', 'phi']]
        x = xc - length * np.sin(phi * np.pi / 180)
        y = yc + length * np.cos(phi * np.pi / 180)
        res[ind]['x'] = x / 10
        res[ind]['y'] = y / 10

    return res


def table_callback(table, data, event, column=False):
    """
    Callback template for tables used together with a
    pn.widget.IntSlider.

    :param table: pn.Panel object of a pd.DataFrame
    :param data: pd.DataFrame storing the data.
    :param event: Slider event returned by param.watch
    :param column: Boolean if true uses horizontal columns to present
        data.
    """
    if column:
        table.object = data.loc[event.new]
    else:
        table.object = data.loc[event.new:event.new]




import json
import strax
import fsspec
import straxen
from numpy import isin
import inspect
from urllib.parse import urlparse, parse_qs
from ast import literal_eval
from functools import lru_cache

from strax.config import OMITTED

export, __all__ = strax.exporter()


def parse_val(val):
    try:
        val = literal_eval(val)
    except ValueError:
        pass
    return val


@export
class URLConfig(strax.Config):
    """Dispatch on URL protocol.
    unrecognized protocol returns identity
    inspired by dasks Dispatch and fsspec fs protocols.
    """
    _LOOKUP = {}
    SCHEME_SEP = '://'
    QUERY_SEP = '?'
    PLUGIN_ATTR_PREFIX = 'plugin.'

    def __init__(self, cache=0, **kwargs):
        """
        :param cache: number of values to keep in cache, 
                      if set to True will cache all values
        :param \**kwargs: additional keyword arguments accepted by strax.Option              
        """
        self.final_type = OMITTED
        super().__init__(**kwargs)
        # Ensure backwards compatibility with Option validation
        # type of the config value can be different from the fetched value.
        if self.type is not OMITTED:
            self.final_type = self.type
            self.type = OMITTED # do not enforce type on the URL
        if cache:
            maxsize = cache if isinstance(cache, int) else None
            self.dispatch = lru_cache(maxsize)(self.dispatch)

    @classmethod
    def register(cls, protocol, func=None):
        """Register dispatch of `func` on urls
        starting with protocol name `protocol` """

        def wrapper(func):
            if isinstance(protocol, tuple):
                for t in protocol:
                    cls.register(t, func)
                return func

            if not isinstance(protocol, str):
                raise ValueError('Protocol name must be a string.')

            if protocol in cls._LOOKUP:
                raise ValueError(f'Protocol with name {protocol} already registered.')
            cls._LOOKUP[protocol] = func    
            return func
        return wrapper(func) if func is not None else wrapper

    def dispatch(self, url, *args, **kwargs):
        """
        Call the corresponding method based on protocol in url.
        chained protocols will be called with the result of the
        previous protocol as input
        overrides are passed to any protocol whos signature can accept them.
        """
        
        # seperate the protocol name from the path
        protocol, _, path =  url.partition(self.SCHEME_SEP)

        # find the corresponding protocol method
        meth = self._LOOKUP.get(protocol, None)
        if meth is None:
            # unrecongnized protocol
            # evaluate as string-literal
            return url

        if self.SCHEME_SEP in path:
            # url contains a nested protocol
            # first call sub-protocol
            arg = self.dispatch(path, **kwargs)
        else:
            # we are at the end of the chain
            # method should be called with path as argument
            arg = path

        # filter kwargs to pass only the kwargs
        #  accepted by the method.
        kwargs = self.filter_kwargs(meth, kwargs)

        return meth(arg, *args, **kwargs)
    
    def split_url_kwargs(self, url):
        """split a url into path and kwargs
        """
        path, _, _ = url.rpartition(self.QUERY_SEP)
        kwargs = {}
        for k,v in parse_qs(urlparse(url).query).items():
            # values of query arguments are evaluated as lists
            # split logic depending on length
            n = len(v)
            if not n:
                kwargs[k] = None
            elif n==1:
                kwargs[k] = parse_val(v[0])
            else:
                kwargs[k] = map(parse_val, v)
        return path, kwargs
    
    @staticmethod
    def filter_kwargs(func, kwargs):
        """Filter out keyword arguments that
            are not in the call signature of func
            and return filtered kwargs dictionary
        """
        params = inspect.signature(func).parameters
        if any([str(p).startswith('**') for p in params.values()]):
            # if func accepts wildcard kwargs, return all
            return kwargs
        return {k:v for k,v in kwargs.items() if k in params}

    def fetch(self, plugin):
        '''override the Config.fetch method
           this is called when the attribute is accessed 
        '''
        # first fetch the user-set value 
        # from the config dictionary
        url = super().fetch(plugin)

        if not isinstance(url, str):
            # if the value is not a string it is evaluated
            # as a literal config and returned as is.
            return url
        
        if self.SCHEME_SEP not in url:
            # no protocol in the url so its evaluated 
            # as string-literal config and returned as is
            return url

        # sperate out the query part of the URL which 
        # will become the method kwargs
        url, url_kwargs = self.split_url_kwargs(url)

        kwargs = {}
        for k,v in url_kwargs.items():
            if isinstance(v, str) and v.startswith(self.PLUGIN_ATTR_PREFIX):
                # kwarg is referring to a plugin attribute, lets fetch it
                kwargs[k] = getattr(plugin, v[len(self.PLUGIN_ATTR_PREFIX):], v)
            else:
                # kwarg is a literal, add its value to the kwargs dict
                kwargs[k] = v
        
        return self.dispatch(url, **kwargs)
 
@strax.URLConfig.register('cmt')
def get_correction(name, run_id=None, version='ONLINE', detector='nt', **kwargs):
    '''Get value for name from CMT
    '''
    if run_id is None:
        raise ValueError('Attempting to fetch a correction without a run id.')
    return straxen.get_correction_from_cmt(run_id, ( name, version, detector=='nt'))

@strax.URLConfig.register('resource')
def get_resource(name, fmt='text', **kwargs):
    '''Fetch a straxen resource
    '''
    return straxen.get_resource(name, fmt=fmt)


@strax.URLConfig.register('fsspec')
def read_file(path, **kwargs):
    '''Support fetching files from arbitrary filesystems
    '''
    with fsspec.open(path, **kwargs) as f:
        content = f.read()
    return content

@strax.URLConfig.register('json')
def read_json(content, **kwargs):
    ''' Load json string as a python object
    '''
    return json.loads(content)

@strax.URLConfig.register('take')
def get_key(container, take=None, **kwargs):
    ''' return a single element of a container
    '''
    if take is None:
        return container
    return container[take]

@strax.URLConfig.register('format')
def format_arg(arg, **kwargs):
    '''apply pythons builtin format function to a string
    '''
    return arg.format(**kwargs)
    


import numba
import ctypes
import strax
from scipy.special.cython_special import loggamma, betainc, gammaln

export, __all__ = strax.exporter()
# Numba versions of scipy functions:
# Based on http://numba.pydata.org/numba-doc/latest/extending/high-level.html but
# omitting the address step.
# Unfortunately this approach does not allow to cache the resulting numba functions.
#  Any ideas in this regard would be welcome.
# All subsequent function will treat their inputs as 64bit floats:
double = ctypes.c_double

functype = ctypes.CFUNCTYPE(double, double)
gammaln_float64 = functype(gammaln)
@export
@numba.njit
def numba_gammaln(x):
    return gammaln_float64(x)


functype = ctypes.CFUNCTYPE(double, double, double, double)
betainc_float64 = functype(betainc)
@export
@numba.njit
def numba_betainc(x1, x2, x3):
    return betainc_float64(x1, x2, x3)


functype = ctypes.CFUNCTYPE(double, double)
loggamma_float64 = functype(loggamma)
@export
@numba.njit
def numba_loggamma(x):
    return loggamma_float64(x)



import os
import re
import typing
import socket
from tqdm import tqdm
from copy import deepcopy
import strax
from .rucio import key_to_rucio_did, RucioLocalBackend
import warnings

try:
    import utilix
except (RuntimeError, FileNotFoundError):
    # We might be on a travis job
    pass
from straxen import uconfig

export, __all__ = strax.exporter()


@export
class RunDB(strax.StorageFrontend):
    """
    Frontend that searches RunDB MongoDB for data.
    """
    # Dict of alias used in rundb: regex on hostname
    hosts = {
        'dali': r'^dali.*rcc.*',
    }

    provide_run_metadata = True

    def __init__(self,
                 minimum_run_number=7157,
                 maximum_run_number=None,
                 runid_field='name',
                 local_only=False,
                 new_data_path=None,
                 reader_ini_name_is_mode=False,
                 rucio_path=None,
                 mongo_url=None,
                 mongo_user=None,
                 mongo_password=None,
                 mongo_database=None,
                 *args, **kwargs):
        """
        :param minimum_run_number: Lowest number to consider
        :param maximum_run_number: Highest number to consider. When None
            (the default) consider all runs that are higher than the
            minimum_run_number.
        :param runid_field: Rundb field to which strax's run_id concept
            corresponds. Can be either
            - 'name': values must be strings, for XENON1T
            - 'number': values must be ints, for XENONnT DAQ tests
        :param local_only: Do not show data as available if it would
            have to be downloaded from a remote location.
        :param new_data_path: Path where new files are to be written.
            Defaults to None: do not write new data
            New files will be registered in the runs db!
            TODO: register under hostname alias (e.g. 'dali')
        :param reader_ini_name_is_mode: If True, will overwrite the
            'mode' field with 'reader.ini.name'.
        :param rucio_path: What is the base path where Rucio is mounted
        :param mongo_url: URL to Mongo runs database (excl auth)
        :param mongo_user: user to Mongo runs database
        :param mongo_password: password to Mongo runs database
        :param mongo_database: database name of Mongo runs database

        Other (kw)args are passed to StorageFrontend.__init__
        """
        super().__init__(*args, **kwargs)
        self.local_only = local_only
        self.new_data_path = new_data_path
        self.reader_ini_name_is_mode = reader_ini_name_is_mode
        self.minimum_run_number = minimum_run_number
        self.maximum_run_number = maximum_run_number
        self.rucio_path = rucio_path
        if self.new_data_path is None:
            self.readonly = True
        self.runid_field = runid_field

        if self.runid_field not in ['name', 'number']:
            raise ValueError("Unrecognized runid_field option %s" % self.runid_field)

        self.hostname = socket.getfqdn()
        if not self.readonly and self.hostname.endswith('xenon.local'):
            # We want admin access to start writing data!
            mongo_url = uconfig.get('rundb_admin', 'mongo_rdb_url')
            mongo_user = uconfig.get('rundb_admin', 'mongo_rdb_username')
            mongo_password = uconfig.get('rundb_admin', 'mongo_rdb_password')
            mongo_database = uconfig.get('rundb_admin', 'mongo_rdb_database')

        # setup mongo kwargs...
        # utilix.rundb.pymongo_collection will take the following variables as kwargs
        #     url: mongo url, including auth
        #     user: the user
        #     password: the password for the above user
        #     database: the mongo database name
        # finally, it takes the collection name as an arg (not a kwarg).
        # if no collection arg is passed, it defaults to the runsDB collection
        # See github.com/XENONnT/utilix/blob/master/utilix/rundb.py for more details
        mongo_kwargs = {'url': mongo_url,
                        'user': mongo_user,
                        'password': mongo_password,
                        'database': mongo_database}
        self.collection = utilix.rundb.xent_collection(**mongo_kwargs)

        # Do not delete the client!
        self.client = self.collection.database.client

        self.backends = [
            strax.FileSytemBackend(),
        ]

        # Construct mongo query for runs with available data.
        # This depends on the machine you're running on.
        self.available_query = [{'host': self.hostname}]

        # Go through known host aliases
        for host_alias, regex in self.hosts.items():
            if re.match(regex, self.hostname):
                self.available_query.append({'host': host_alias})

        if self.rucio_path is not None:
            self.backends.append(RucioLocalBackend(self.rucio_path))
            # When querying for rucio, add that it should be dali-userdisk
            self.available_query.append({'host': 'rucio-catalogue',
                                         'location': 'UC_DALI_USERDISK',
                                         'status': 'transferred',
                                         })

    def _data_query(self, key):
        """Return MongoDB query for data field matching key"""
        return {
            'data': {
                '$elemMatch': {
                    'type': key.data_type,
                    # TODO remove the meta.lineage since this doc
                    #  entry is deprecated.
                    '$and': [{'$or': [
                        {'meta.lineage': key.lineage},
                        {'did':
                             {'$regex':
                                  f'/*{key.data_type}-{key.lineage_hash}'
                              },
                         },
                    ]},
                        {'$or': self.available_query}]}}}

    def _find(self, key: strax.DataKey,
              write, allow_incomplete, fuzzy_for, fuzzy_for_options):
        if key.run_id.startswith('_'):
            # Superruns are currently not supprorted..
            raise strax.DataNotAvailable
        
        if fuzzy_for or fuzzy_for_options:
            warnings.warn("Can't do fuzzy with RunDB yet. Only returning exact matches")

        # Check if the run exists
        if self.runid_field == 'name':
            run_query = {'name': str(key.run_id)}
        else:
            run_query = {'number': int(key.run_id)}

        # Check that we are in rucio backend
        if self.rucio_path is not None:
            rucio_key = key_to_rucio_did(key)
            rucio_available_query = self.available_query[-1]
            dq = {
                'data': {
                    '$elemMatch': {
                        'type': key.data_type,
                        'did': rucio_key,
                        **rucio_available_query,
                    },
                }}
            doc = self.collection.find_one({**run_query,
                                            **dq,
                                            },
                                           projection=dq)
            if doc is not None:
                datum = doc['data'][0]
                error_message = f'Expected {rucio_key} got data on {datum["location"]}'
                assert datum.get('did', '') == rucio_key, error_message
                backend_name = 'RucioLocalBackend'
                backend_key = key_to_rucio_did(key)
                return backend_name, backend_key

        dq = self._data_query(key)
        doc = self.collection.find_one({**run_query, **dq}, projection=dq)

        if doc is None:
            # Data was not found
            if not write:
                raise strax.DataNotAvailable

            output_path = os.path.join(self.new_data_path, str(key))

            if self.new_data_path is not None:
                doc = self.collection.find_one(run_query, projection={'_id'})
                if not doc:
                    raise ValueError(f"Attempt to register new data for"
                                     f" non-existing run {key.run_id}")
                self.collection.find_one_and_update(
                    {'_id': doc['_id']},
                    {'$push': {'data': {
                        'location': output_path,
                        'host': self.hostname,
                        'type': key.data_type,
                        'protocol': strax.FileSytemBackend.__name__,
                        # TODO: duplication with metadata stuff elsewhere?
                        'meta': {'lineage': key.lineage}
                    }}})

            return (strax.FileSytemBackend.__name__,
                    output_path)
        datum = doc['data'][0]

        if datum['host'] == 'rucio-catalogue':
            # TODO this is due to a bad query in _data_query. We aren't rucio.
            raise strax.DataNotAvailable

        if write and not self._can_overwrite(key):
            raise strax.DataExistsError(at=datum['location'])

        return datum['protocol'], datum['location']

    def find_several(self, keys: typing.List[strax.DataKey], **kwargs):
        if kwargs.get('fuzzy_for', False) or kwargs.get('fuzzy_for_options', False):
            warnings.warn("Can't do fuzzy with RunDB yet. Only returning exact matches")
        if not len(keys):
            return []
        if not len(set([k.lineage_hash for k in keys])) == 1:
            raise ValueError("find_several keys must have same lineage")
        if not len(set([k.data_type for k in keys])) == 1:
            raise ValueError("find_several keys must have same data type")
        keys = list(keys)  # Context used to pass a set

        if self.runid_field == 'name':
            run_query = {'name': {'$in': [key.run_id for key in keys]}}
        else:
            run_query = {f'{self.runid_field}': {'$in': [int(key.run_id) for key in keys if not key.run_id.startswith('_')]}}
        dq = self._data_query(keys[0])

        # dict.copy is sometimes not sufficient for nested dictionary
        projection = deepcopy(dq)
        projection.update({
            k: True
            for k in f'name number'.split()})

        results_dict = dict()
        for doc in self.collection.find(
                {**run_query, **dq}, projection=projection):
            # If you get a key error here there might be something off with the
            # projection
            datum = doc['data'][0]

            if self.runid_field == 'name':
                dk = doc['name']
            else:
                dk = f'{doc["number"]:06}'
            try:
                results_dict[dk] = datum['protocol'], datum['location']
            except KeyError as e:
                raise KeyError(f'Queries failed\n{run_query}\n{dq}\n{doc}') from e
        return [results_dict.get(k.run_id, False)
                for k in keys]

    def _scan_runs(self, store_fields):
        query = self.number_query()
        projection = strax.to_str_tuple(list(store_fields))
        # Replace fields by their subfields if requested only take the most
        # "specific" projection
        projection = [f1 for f1 in projection
                      if not any([f2.startswith(f1 + ".") for f2 in projection])]
        cursor = self.collection.find(
            filter=query,
            projection=projection)
        for doc in strax.utils.tqdm(
                cursor, desc='Fetching run info from MongoDB',
                total=cursor.count()):
            del doc['_id']
            if self.reader_ini_name_is_mode:
                doc['mode'] = \
                    doc.get('reader', {}).get('ini', {}).get('name', '')
            yield doc

    def run_metadata(self, run_id, projection=None):
        if run_id.startswith('_'):
            # Superruns are currently not supported..
            raise strax.DataNotAvailable
        
        if self.runid_field == 'name':
            run_id = str(run_id)
        else:
            run_id = int(run_id)
        if isinstance(projection, str):
            projection = {projection: 1}
        elif isinstance(projection, (list, tuple)):
            projection = {x: 1 for x in projection}

        doc = self.collection.find_one(
            {self.runid_field: run_id},
            projection=projection)
        if doc is None:
            raise strax.DataNotAvailable
        if self.reader_ini_name_is_mode:
            doc['mode'] = doc.get('reader', {}).get('ini', {}).get('name', '')
        return doc

    def number_query(self):
        q_number = {}
        if self.minimum_run_number:
            q_number = {'$gt': self.minimum_run_number}
        if self.maximum_run_number is not None:
            q_number.update({'$lt': self.maximum_run_number})
        if q_number:
            return {'number': q_number}
        return {}



import strax
import straxen
import tarfile
import io
import os
from warnings import warn
from os import environ as os_environ
from straxen import aux_repo, pax_file
from pandas import DataFrame
from immutabledict import immutabledict
import numpy as np


export, __all__ = strax.exporter()


nt_test_run_id = '012882'
test_run_id_1T = '180423_1021'

testing_config_1T = dict(
    hev_gain_model=('1T_to_pe_placeholder', False),
    gain_model=('1T_to_pe_placeholder', False),
    elife_conf=('elife_constant', 1e6),
    electron_drift_velocity=("electron_drift_velocity_constant", 1e-4),
    electron_drift_time_gate=("electron_drift_time_gate_constant", 1700),
)

# Let's make a dummy map for NVeto
_nveto_pmt_dummy = {'channel': list(range(2000, 2120)),
                    'x': list(range(120)),
                    'y': list(range(120)),
                    'z': list(range(120)),
                    }
_nveto_pmt_dummy_df = DataFrame(_nveto_pmt_dummy)

# Some configs are better obtained from the strax_auxiliary_files repo.
# Let's use small files, we don't want to spend a lot of time downloading
# some file.
_testing_config_nT = dict(
    nn_architecture=
    aux_repo + 'f0df03e1f45b5bdd9be364c5caefdaf3c74e044e/fax_files/mlp_model.json',
    nn_weights=
    aux_repo + 'f0df03e1f45b5bdd9be364c5caefdaf3c74e044e/fax_files/mlp_model.h5',
    gain_model=("to_pe_placeholder", True),
    s2_xy_correction_map=pax_file('XENON1T_s2_xy_ly_SR0_24Feb2017.json'),
    elife_conf=('elife_constant', 1e6),
    baseline_samples_nv=10,
    fdc_map=pax_file('XENON1T_FDC_SR0_data_driven_3d_correction_tf_nn_v0.json.gz'),
    gain_model_nv=("adc_nv", True),
    gain_model_mv=("adc_mv", True),
    nveto_pmt_position_map=_nveto_pmt_dummy,
    s1_xyz_correction_map=pax_file('XENON1T_s1_xyz_lce_true_kr83m_SR0_pax-680_fdc-3d_v0.json'),
    electron_drift_velocity=("electron_drift_velocity_constant", 1e-4),
    s1_aft_map=aux_repo + '023cb8caf2008b289664b0fefc36b1cebb45bbe4/strax_files/s1_aft_UNITY_xyz_XENONnT.json',  # noqa
    s2_optical_map=aux_repo + '9891ee7a52fa00e541480c45ab7a1c9a72fcffcc/strax_files/XENONnT_s2_xy_unity_patterns.json.gz',  # noqa
    s1_optical_map=aux_repo + '9891ee7a52fa00e541480c45ab7a1c9a72fcffcc/strax_files/XENONnT_s1_xyz_unity_patterns.json.gz',  # noqa
    electron_drift_time_gate=("electron_drift_time_gate_constant", 2700),
    hit_min_amplitude='pmt_commissioning_initial',
    hit_min_amplitude_nv=20,
    hit_min_amplitude_mv=80,
    hit_min_amplitude_he='pmt_commissioning_initial_he'
)


@export
def download_test_data(test_data='https://raw.githubusercontent.com/XENONnT/strax_auxiliary_files/353b2c60a01e96f67e4ba544ce284bd91241964d/strax_files/strax_test_data_straxv1.1.0.tar',  #  noqa
                       ):
    """Downloads strax test data to strax_test_data in the current directory"""
    blob = straxen.common.get_resource(test_data, fmt='binary')
    f = io.BytesIO(blob)
    tf = tarfile.open(fileobj=f)
    tf.extractall()


@export
def _overwrite_testing_function_file(function_file):
    """For testing purposes allow this function file to be loaded from HOME/testing_folder"""
    if not _is_on_pytest():
        # If we are not on a pytest, never try using a local file.
        return function_file

    home = os.environ.get('HOME')
    if home is None:
        # Impossible to load from non-existent folder
        return function_file

    testing_file = os.path.join(home, function_file)

    if os.path.exists(testing_file):
        # For testing purposes allow loading from 'home/testing_folder'
        warn(f'Using local function: {function_file} from {testing_file}! '
             f'If you are not integrated testing on github you should '
             f'absolutely remove this file. (See #559)')
        function_file = testing_file

    return function_file


@export
def _is_on_pytest():
    """Check if we are on a pytest"""
    return 'PYTEST_CURRENT_TEST' in os_environ


def nt_test_context(target_context='xenonnt_online',
                    deregister = ('peak_veto_tags', 'events_tagged'),
                    **kwargs):
    st = getattr(straxen.contexts, target_context)(**kwargs)
    st._plugin_class_registry['raw_records'].__version__ = "MOCKTESTDATA"  # noqa
    st.storage = [strax.DataDirectory('./strax_test_data')]
    download_test_data('https://raw.githubusercontent.com/XENONnT/strax_auxiliary_files/1d3706d4b47cbd23b5cae66d5e258bb84487ad01/strax_files/012882-raw_records-z7q2d2ye2t.tar')  # noqa
    assert st.is_stored(nt_test_run_id, 'raw_records'), os.listdir(st.storage[-1].path)

    to_remove = list(deregister)
    if not straxen.utilix_is_configured(warning_message=False):
        st.set_config(_testing_config_nT)
        to_remove += 'peak_positions_mlp peak_positions_cnn peak_positions_gcn s2_recon_pos_diff'.split()  # noqa
        # TODO The test data for this plugin doesn't work
        to_remove += ['event_pattern_fit']
        st.set_config({'gain_model': ("to_pe_placeholder", True)})
        st.register(straxen.PeakPositions1T)
    for plugin in to_remove:
        del st._plugin_class_registry[plugin]
    return st


@strax.takes_config(
    strax.Option('secret_time_offset', default=0, track=False),
    strax.Option('recs_per_chunk', default=10, track=False),
    strax.Option('n_chunks', default=2, track=False,
                 help='Number of chunks for the dummy raw records we are writing here'),
    strax.Option('channel_map', track=False, type=immutabledict,
                 help="frozendict mapping subdetector to (min, max) "
                      "channel number.")
)
class DummyRawRecords(strax.Plugin):
    """
    Provide dummy raw records for the mayor raw_record types
    """
    provides = ('raw_records',
                'raw_records_he',
                'raw_records_nv',
                'raw_records_aqmon',
                'raw_records_aux_mv',
                'raw_records_mv'
                )
    parallel = 'process'
    depends_on = tuple()
    data_kind = immutabledict(zip(provides, provides))
    rechunk_on_save = False
    dtype = {p: strax.raw_record_dtype() for p in provides}

    def setup(self):
        self.channel_map_keys = {'he': 'he',
                                 'nv': 'nveto',
                                 'aqmon': 'aqmon',
                                 'aux_mv': 'aux_mv',
                                 's_mv': 'mv',
                                 }  # s_mv otherwise same as aux in endswith

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < self.config['n_chunks']

    def compute(self, chunk_i):
        t0 = chunk_i + self.config['secret_time_offset']
        if chunk_i < self.config['n_chunks'] - 1:
            # One filled chunk
            r = np.zeros(self.config['recs_per_chunk'], self.dtype['raw_records'])
            r['time'] = t0
            r['length'] = r['dt'] = 1
            r['channel'] = np.arange(len(r))
        else:
            # One empty chunk
            r = np.zeros(0, self.dtype['raw_records'])

        res = {}
        for p in self.provides:
            rr = np.copy(r)
            # Add detector specific channel offset:
            for key, channel_key in self.channel_map_keys.items():
                if channel_key not in self.config['channel_map']:
                    # Channel map for 1T is different.
                    continue
                if p.endswith(key):
                    s, e = self.config['channel_map'][channel_key]
                    rr['channel'] += s
            res[p] = self.chunk(start=t0, end=t0 + 1, data=rr, data_type=p)
        return res



import strax
import straxen

import numpy as np

import hypothesis
from hypothesis import given, settings
import hypothesis.strategies as hst
import hypothesis.extra.numpy as hnp


@hst.composite
def create_disjoint_intervals(draw,
                              dtype,
                              n_intervals=10,
                              dt=1,
                              time_range=(0, 100),
                              channel_range=(2000, 2119),
                              length_range=(1, 1), ):
    """
    Function which generates a hypothesis strategy for a fixed number
    of disjoint intervals

    :param dtype: Can be any strax-like dtype either with endtime or
        dt and length field.
    :param n_intervals: How many disjoint intervals should be returned.
    :param dt: Sampling field, only needed for length + dt fields.
    :param time_range: Time range in which random numbers will be
        generated.
    :param channel_range: Range of channels for which the disjoint
        intervals will be generated. For a single channel set min/max
        equal.
    :param length_range: Range how long time intervals can be.
    :return: hypothesis strategy which can be used in @given

    Note:
        You can use create_disjoint_intervals().example() to see an
        example.
        If you do not want to specify the bounds for any of the "_range"
        parameters set the corresponding bound to None.

        Somehow hypothesis complains that the creation of these events
        takes too long ~2 s for 50 intervals. You can disable the
        corresponding healt checks via:" @settings(
        suppress_health_check=[hypothesis.HealthCheck.large_base_example,
         hypothesis.HealthCheck.too_slow])"
    """
    n = 0

    if not hasattr(dtype, 'fields'):
        # Convert dtype into numpy dtype
        dtype = np.dtype(dtype)

    is_dt = True
    if 'endtime' in dtype.fields:
        # Check whether interval uses dt fields or endtime
        is_dt = False

    stratgey_example = np.zeros(n_intervals, dtype)
    if is_dt:
        stratgey_example['dt'] = dt

    while n < n_intervals:
        # Create interval values:
        time = draw(hst.integers(*time_range))
        channel = draw(hst.integers(*channel_range))
        length = draw(hst.integers(*length_range))

        # Check if objects are disjoint:
        if _test_disjoint(stratgey_example[:n], time, length, channel, dt):
            stratgey_example[n]['time'] = time
            stratgey_example[n]['channel'] = channel
            if is_dt:
                stratgey_example[n]['length'] = length
            else:
                stratgey_example[n]['endtime'] = time + int(length * dt)
            n += 1
    return stratgey_example


def _test_disjoint(intervals, time, length, channel, dt):
    int_ch = intervals[intervals['channel'] == channel]
    if not len(int_ch):
        # No interval in the given channel yet
        return True

    endtime = strax.endtime(int_ch)
    m = (int_ch['time'] <= time) & (time < endtime)
    edt = time + length * dt
    m |= (int_ch['time'] <= edt) & (edt < endtime)
    if np.any(m):
        # Found an overlap:
        return False
    else:
        return True


@settings(suppress_health_check=[hypothesis.HealthCheck.large_base_example,
                                 hypothesis.HealthCheck.too_slow],
          deadline=None)
@given(create_disjoint_intervals(strax.hitlet_dtype(),
                                 n_intervals=7,
                                 dt=1,
                                 time_range=(0, 15),
                                 channel_range=(2000, 2010),
                                 length_range=(1, 4), ),
       hst.integers(1, 3),
       )
def test_nveto_event_building(hitlets,
                              coincidence):
    """
    In this test we test the code of
    straxen.plugins.veto_evnets.find_veto_events
    """
    hitlets = strax.sort_by_time(hitlets)

    event_intervals = straxen.plugins.nveto_recorder.find_coincidence(hitlets,
                                                                      coincidence,
                                                                      300)

    mes = 'Found overlapping events returned by "coincidence".'
    assert np.all(event_intervals['endtime'][:-1] - event_intervals['time'][1:] < 0), mes

    # Get hits which touch the event window, this can lead to ambiguities
    # which we will solve subsequently.
    hitlets_ids_in_event = strax.touching_windows(hitlets, event_intervals)
    # First check for empty events, since ambiguity check will merge intervals:
    mes = f'Found an empty event without any hitlets: {hitlets_ids_in_event}.'
    assert np.all(np.diff(hitlets_ids_in_event) != 0), mes

    # Solve ambiguities (merge overlapping intervals)
    interval_truth = _test_ambiguity(hitlets_ids_in_event)
    hitlets_ids_in_event = straxen.plugins.veto_events._solve_ambiguity(hitlets_ids_in_event)

    mes = f'Found ambigious event for {hitlets_ids_in_event} with turth {interval_truth}'
    assert np.all(hitlets_ids_in_event == interval_truth), mes

    # Check if events satisfy the coincidence requirement:
    mes = f'Found an event with less than 3 hitelts. {hitlets_ids_in_event}'
    assert np.all(np.diff(hitlets_ids_in_event) >= coincidence), mes


def _test_ambiguity(hitlets_ids_in_event):
    """
    Returns overlap free intervals for ambiguity check.
    """
    res = []
    start, end = hitlets_ids_in_event[0]
    for ids in hitlets_ids_in_event[1:]:
        s, e = ids
        if s < end:
            # Overlapping event:
            end = e
        else:
            # New event:
            res.append([start, end])
            start = s
            end = e
    # Add last interval:
    res.append([start, end])
    return res


@settings(suppress_health_check=[hypothesis.HealthCheck.large_base_example,
                                 hypothesis.HealthCheck.too_slow],
          deadline=None)
@given(create_disjoint_intervals(strax.hitlet_dtype(),
                                 n_intervals=7,
                                 dt=1,
                                 time_range=(0, 15),
                                 channel_range=(2000, 2010),
                                 length_range=(1, 4), ),
       hnp.arrays(np.float32, elements=hst.floats(-1, 10, width=32), shape=7),
       )
def test_nveto_event_plugin(hitlets, area):
    hitlets['area'] = area
    hitlets = strax.sort_by_time(hitlets)
    events, hitlets_ids_in_event = straxen.find_veto_events(hitlets,
                                                            3,
                                                            300,
                                                            0)

    straxen.plugins.veto_events.compute_nveto_event_properties(events,
                                                               hitlets,
                                                               hitlets_ids_in_event,
                                                               start_channel=2000)
    # Test some of the parameters:
    for e, hit_ids in zip(events, hitlets_ids_in_event):
        hits = hitlets[hit_ids[0]:hit_ids[1]]

        assert e['time'] == np.min(hits['time']), f'Event start is wrong (hit_ids: hit_ids)'
        assert e['endtime'] == np.max(strax.endtime(hits)), f'Event end is wrong (hit_ids: hit_ids)'
        assert np.isclose(e['area'], np.sum(hits['area'])), f'Event area is wrong for {e["area"]}, {hits["area"]}'
        mes = f'Event n_contributing_pmt is wrong for {e["n_contributing_pmt"]}, {hits["channel"]}'
        assert e['n_contributing_pmt'] == len(np.unique(hits['channel'])), mes
        assert e['n_hits'] == len(hits), f'Event n_hits is wrong for {e["n_hits"]}, {hits}'

    # -----------------------
    # Check if updated events
    # have the correct boundaries:
    # -----------------------
    if len(events) > 1:
        mes = f'Updated event boundaries overlap! {events}'
        assert (events['endtime'][:-1] - events['time'][1:]) < 0, mes

    split_hitlets = strax.split_by_containment(hitlets, events)

    for sbc_hitlets, tw_hitlet_id in zip(split_hitlets, hitlets_ids_in_event):
        h = hitlets[tw_hitlet_id[0]:tw_hitlet_id[1]]
        mes = ('Touching windows and split_by_containment yield different hitlets'
               ' after updating the event boundaries. This should not have happened.')
        assert np.all(sbc_hitlets == h), mes

    # Test event positions:
    try:
        npmt_pos = straxen.get_resource('nveto_pmt_position.csv', fmt='csv')
        npmt_pos = npmt_pos.to_records(index=False)
    except FileNotFoundError:
        npmt_pos = np.ones(120, dtype=[('x', np.float64),
                                       ('y', np.float64),
                                       ('z', np.float64)])

    events_angle = np.zeros(len(events),
                            dtype=straxen.plugins.veto_events.veto_event_positions_dtype())

    straxen.plugins.veto_events.compute_positions(events_angle,
                                                  events,
                                                  split_hitlets,
                                                  npmt_pos,
                                                  start_channel=2000)

    angle = straxen.plugins.veto_events.get_average_angle(split_hitlets,
                                                          npmt_pos,
                                                          start_channel=2000)
    # Compute truth angles:
    truth_angle = np.angle(events_angle['pos_x']+events_angle['pos_y']*1j)
    # Replace not defined angles, into zeros to match np.angles return
    # and to simplify comparison
    m = (events_angle['pos_x'] == 0) & (events_angle['pos_y'] == 0)
    angle[m] = 0

    # Fixing +2pi issue and np.angle [-180, 180] and [0, 360) convention
    # issue.
    angle = angle % (2*np.pi)
    truth_angle = truth_angle % (2*np.pi)

    # Sometimes it may happen due to numerical precision that one angle is slightly
    # larger than 2 pi while the other is slightly smaller. In that case we have to
    # fix it:
    if np.isclose(angle, 2*np.pi):
        angle -= 2*np.pi
    if np.isclose(truth_angle, 2*np.pi):
        truth_angle -= 2*np.pi

    # Compare angle, also indirectly tests average x/y/z
    mes = f'Event angle did not match expected {truth_angle}, got {angle}.'
    assert np.isclose(angle, truth_angle), mes



import numpy as np
import straxen
import unittest
import requests


class SCInterfaceTest(unittest.TestCase):
    def setUp(self):
        self.resources_available()
        # Simple query test:
        # Query 5 s of data:
        self.start = 1609682275000000000
        # Add micro-second to check if query does not fail if inquery precsion > SC precision
        self.start += 10**6
        self.end = self.start + 5*10**9

    def test_query_sc_values(self):
        """
        Unity test for the SCADAInterface. Query a fixed range and check if 
        return is correct.
        """
        print('Testing SCADAInterface')
        parameters = {'SomeParameter': 'XE1T.CTPC.Board06.Chan011.VMon'}
        df = self.sc.get_scada_values(parameters,
                                      start=self.start,
                                      end=self.end,
                                      every_nth_value=1,
                                      query_type_lab=False, )

        assert df['SomeParameter'][0] // 1 == 1253, 'First values returned is not corrrect.'
        assert np.all(np.isnan(df['SomeParameter'][1:])), 'Subsequent values are not correct.'

        print('Testing forwardfill option:')
        parameters = {'SomeParameter': 'XE1T.CRY_FCV104FMON.PI'}
        df = self.sc.get_scada_values(parameters,
                                      start=self.start,
                                      end=self.end,
                                      fill_gaps='forwardfill',
                                      every_nth_value=1,
                                      query_type_lab=False,)
        assert np.all(np.isclose(df[:4], 2.079859)), 'First four values deviate from queried values.'
        assert np.all(np.isclose(df[4:], 2.117820)), 'Last two values deviate from queried values.'

        print('Testing down sampling and averaging option:')
        parameters = {'SomeParameter': 'XE1T.CRY_TE101_TCRYOBOTT_AI.PI'}
        df_all = self.sc.get_scada_values(parameters,
                                          start=self.start,
                                          end=self.end,
                                          fill_gaps='forwardfill',
                                          every_nth_value=1,
                                          query_type_lab=False, )

        df = self.sc.get_scada_values(parameters,
                                      start=self.start,
                                      end=self.end,
                                      fill_gaps='forwardfill',
                                      down_sampling=True,
                                      every_nth_value=2,
                                      query_type_lab=False,)

        assert np.all(df_all[::2] == df), 'Downsampling did not return the correct values.'

        df = self.sc.get_scada_values(parameters,
                                      start=self.start,
                                      end=self.end,
                                      fill_gaps='forwardfill',
                                      every_nth_value=2,
                                      query_type_lab=False,)

        # Compare average for each second value:
        for ind, i in enumerate([0, 2, 4]):
            is_correct = np.isclose(np.mean(df_all[i:i + 2]), df['SomeParameter'][ind])
            assert is_correct, 'Averaging is incorrect.'

        # Testing lab query type:
        df = self.sc.get_scada_values(parameters,
                                      start=self.start,
                                      end=self.end,
                                      query_type_lab=True,)
        is_sorrect = np.all(df['SomeParameter'] // 1 == -96)
        assert is_sorrect, 'Not all values are correct for query type lab.'

    @staticmethod
    def test_average_scada():
        t = np.arange(0, 100, 10)
        t_t, t_a = straxen.scada._average_scada(t / 1e9, t, 1)
        assert len(t_a) == len(t), 'Scada deleted some of my 10 datapoints!'

    def resources_available(self):
        """
        Exception to skip test if external requirements are not met. Otherwise define 
        Scada interface as self.sc.
        """
        if not straxen.utilix_is_configured('scada','scdata_url',):
            self.skipTest("Cannot test scada since we have no access to xenon secrets.)")

        try:
            self.sc = straxen.SCADAInterface(use_progress_bar=False)
            self.sc.get_new_token()
        except requests.exceptions.SSLError:
            self.skipTest("Cannot reach database since HTTPs certifcate expired.")



import straxen
import pandas
import os
import numpy as np
import strax
from matplotlib.pyplot import clf as plt_clf
from straxen.test_utils import nt_test_context, nt_test_run_id
import unittest


def test_pmt_pos_1t():
    """
    Test if we can get the 1T PMT positions
    """
    pandas.DataFrame(straxen.pmt_positions(True))


def test_pmt_pos_nt():
    """
    Test if we can get the nT PMT positions
    """
    pandas.DataFrame(straxen.pmt_positions(False))


class TestMiniAnalyses(unittest.TestCase):
    """

    """
    # They were added on 25/10/2021 and may be outdated by now
    _expected_test_results = {
        'peak_basics': 40,
        'n_s1': 19,
        'run_live_time': 4.7516763,
        'event_basics': 20,
    }

    @classmethod
    def setUpClass(cls) -> None:
        cls.st = nt_test_context()
        # For al the WF plotting, we might need records, let's make those
        cls.st.make(nt_test_run_id, 'records')
        cls.first_peak = cls.st.get_array(nt_test_run_id, 'peak_basics')[0]
        cls.first_event = cls.st.get_array(nt_test_run_id, 'event_basics')[0]

    def tearDown(self):
        plt_clf()

    def test_target_peaks(self, target='peak_basics', tol=2):
        assert target in self._expected_test_results, f'No expectation for {target}?!'
        data = self.st.get_array(nt_test_run_id, target)
        message = (f'Got more/less data for {target}. If you changed something '
                   f'on {target}, please update the numbers in '
                   f'tests/test_mini_analyses.TestMiniAnalyses._expected_test_results')
        if not straxen.utilix_is_configured():
            # If we do things with dummy maps, things might be slightly different
            tol += 10
        assert np.abs(len(data) - self._expected_test_results[target]) < tol, message

    def test_target_events(self):
        self.test_target_peaks(target='event_basics')

    def test_plot_waveform(self, deep=False):
        self.st.plot_waveform(nt_test_run_id, time_within=self.first_peak, deep=deep)

    def test_plot_waveform_deep(self):
        self.test_plot_waveform(deep=True)

    def test_plot_hit_pattern(self):
        self.st.plot_hit_pattern(nt_test_run_id, time_within=self.first_peak, xenon1t=False)

    def test_plot_records_matrix(self):
        self.st_attr_for_one_peak('plot_records_matrix')

    def test_raw_records_matrix(self):
        self.st_attr_for_one_peak('raw_records_matrix')

    def test_event_display_simple(self):
        plot_all_positions = straxen.utilix_is_configured()
        self.st.event_display_simple(nt_test_run_id,
                                     time_within=self.first_event,
                                     xenon1t=False,
                                     plot_all_positions=plot_all_positions,
                                     )

    def test_single_event_plot(self):
        plot_all_positions = straxen.utilix_is_configured()
        straxen.analyses.event_display.plot_single_event(
            self.st,
            nt_test_run_id,
            events=self.st.get_array(nt_test_run_id, 'events'),
            event_number=self.first_event['event_number'],
            xenon1t=False,
            plot_all_positions=plot_all_positions,
        )

    def test_event_display_interactive(self):
        self.st.event_display_interactive(nt_test_run_id,
                                          time_within=self.first_event,
                                          xenon1t=False,
                                          )

    def test_plot_peaks_aft_histogram(self):
        self.st.plot_peaks_aft_histogram(nt_test_run_id)

    def test_event_scatter(self):
        self.st.event_scatter(nt_test_run_id)

    def test_energy_spectrum(self):
        self.st.plot_energy_spectrum(nt_test_run_id)

    def test_peak_classification(self):
        self.st.plot_peak_classification(nt_test_run_id)

    def st_attr_for_one_peak(self, function_name):
        f = getattr(self.st, function_name)
        f(nt_test_run_id, time_within=self.first_peak)

    def test_waveform_display(self):
        self.st_attr_for_one_peak('waveform_display')

    def test_hvdisp_plot_pmt_pattern(self):
        self.st_attr_for_one_peak('hvdisp_plot_pmt_pattern')

    def test_hvdisp_plot_peak_waveforms(self):
        self.st_attr_for_one_peak('hvdisp_plot_peak_waveforms')

    def test_plot_pulses_tpc(self):
        self.st.plot_pulses_tpc(nt_test_run_id, max_plots=2, plot_hits=True,
                                ignore_time_warning=True)

    def test_plot_pulses_mv(self):
        self.st.plot_pulses_mv(nt_test_run_id, max_plots=2, plot_hits=True,
                               ignore_time_warning=True)

    def test_plot_pulses_nv(self):
        self.st.plot_pulses_nv(nt_test_run_id, max_plots=2, plot_hits=True,
                               ignore_time_warning=True)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_event_display(self):
        self.st.event_display(nt_test_run_id, time_within=self.first_event)

    def test_calc_livetime(self):
        try:
            live_time = straxen.get_livetime_sec(self.st, nt_test_run_id)
        except strax.RunMetadataNotAvailable:
            things = self.st.get_array(nt_test_run_id, 'peaks')
            live_time = straxen.get_livetime_sec(self.st, nt_test_run_id, things=things)
        assertion_statement = "Live-time calculation is wrong"
        assert live_time == self._expected_test_results['run_live_time'], assertion_statement

    def test_df_wiki(self):
        df = self.st.get_df(nt_test_run_id, 'peak_basics')
        straxen.dataframe_to_wiki(df)

    def test_interactive_display(self):
        fig = self.st.event_display_interactive(nt_test_run_id,
                                                time_within=self.first_event,
                                                xenon1t=False,
                                                plot_record_matrix=True,
                                                )
        save_as = 'test_display.html'
        fig.save(save_as)
        assert os.path.exists(save_as)
        os.remove(save_as)
        assert not os.path.exists(save_as), f'Should have removed {save_as}'

    def test_selector(self):
        from straxen.analyses.bokeh_waveform_plot import DataSelectionHist
        p = self.st.get_array(nt_test_run_id, 'peak_basics')
        ds = DataSelectionHist('ds')
        fig = ds.histogram2d(p,
                             p['area'],
                             p['area'],
                             bins=50,
                             hist_range=((0, 200), (0, 2000)),
                             log_color_scale=True,
                             clim=(10, None),
                             undeflow_color='white')

        import bokeh.plotting as bklt
        save_as = 'test_data_selector.html'
        bklt.save(fig, save_as)
        assert os.path.exists(save_as)
        os.remove(save_as)
        assert not os.path.exists(save_as), f'Should have removed {save_as}'

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nt_daq_plot(self):
        self.st.daq_plot(nt_test_run_id,
                         time_within=self.first_peak,
                         vmin=0.1,
                         vmax=1,
                         )

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nt_daq_plot_grouped(self):
        self.st.plot_records_matrix(nt_test_run_id,
                                    time_within=self.first_peak,
                                    vmin=0.1,
                                    vmax=1,
                                    group_by='ADC ID',
                                    )

    def test_records_matrix_downsample(self):
        self.st.records_matrix(nt_test_run_id,
                               time_within=self.first_event,
                               max_samples=20
                               )

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_load_corrected_positions(self):
        self.st.load_corrected_positions(nt_test_run_id, time_within=self.first_peak)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nv_event_display(self):
        self.st.make(nt_test_run_id, 'events_nv')
        self.st.make(nt_test_run_id, 'event_positions_nv')
        with self.assertRaises(ValueError):
            self.st.plot_nveto_event_display(nt_test_run_id, time_within=self.first_peak)


def test_plots():
    """Make some plots"""
    c = np.ones(straxen.n_tpc_pmts)
    straxen.plot_pmts(c)
    straxen.plot_pmts(c, log_scale=True)



import unittest
import strax
from strax.testutils import Records, Peaks
import straxen
import os
import shutil
import tempfile
import pymongo
import datetime


def mongo_uri_not_set():
    return 'TEST_MONGO_URI' not in os.environ


class TestRunDBFrontend(unittest.TestCase):
    """
    Test the saving behavior of the context with the straxen.RunDB

    Requires write access to some pymongo server, the URI of witch is to be set
    as an environment variable under:

        TEST_MONGO_URI

    At the moment this is just an empty database but you can also use some free
    ATLAS mongo server.
    """
    _run_test = True

    def setUp(self):
        # Just to make sure we are running some mongo server, see test-class docstring
        if 'TEST_MONGO_URI' not in os.environ:
            return
        self.test_run_ids = ['0', '1']
        self.all_targets = ('peaks', 'records')

        uri = os.environ.get('TEST_MONGO_URI')
        db_name = 'test_rundb'
        self.collection_name = 'test_rundb_coll'
        client = pymongo.MongoClient(uri)
        self.database = client[db_name]
        collection = self.database[self.collection_name]
        self.path = os.path.join(tempfile.gettempdir(), 'strax_data')
        assert self.collection_name not in self.database.list_collection_names()

        if not straxen.utilix_is_configured():
            # Bit of an ugly hack but there is no way to get around this
            # function even though we don't need it
            straxen.rundb.utilix.rundb.xent_collection = lambda *args, **kwargs: collection

        self.rundb_sf = straxen.RunDB(readonly=False,
                                      runid_field='number',
                                      new_data_path=self.path,
                                      minimum_run_number=-1,
                                      )
        self.rundb_sf.client = client
        self.rundb_sf.collection = collection

        self.st = strax.Context(register=[Records, Peaks],
                                storage=[self.rundb_sf],
                                use_per_run_defaults=False,
                                config=dict(bonus_area=0),
                                )
        for run_id in self.test_run_ids:
            collection.insert_one(_rundoc_format(run_id))
        assert not self.is_all_targets_stored

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    def tearDown(self):
        self.database[self.collection_name].drop()
        if os.path.exists(self.path):
            print(f'rm {self.path}')
            shutil.rmtree(self.path)

    @property
    def is_all_targets_stored(self) -> bool:
        """This should always be False as one of the targets (records) is not stored in mongo"""
        return all([all(
            [self.st.is_stored(r, t) for t in self.all_targets])
            for r in self.test_run_ids])

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    def test_finding_runs(self):
        rdb = self.rundb_sf
        col = self.database[self.collection_name]
        assert col.find_one() is not None
        query = rdb.number_query()
        assert col.find_one(query) is not None
        runs = self.st.select_runs()
        assert len(runs) == len(self.test_run_ids)

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    def test_write_and_load(self):
        assert not self.is_all_targets_stored

        # Make ALL the data
        # NB: the context writes to ALL the storage frontends that are susceptible
        for t in self.all_targets:
            self.st.make(self.test_run_ids, t)

        for r in self.test_run_ids:
            print(self.st.available_for_run(r))
        assert self.is_all_targets_stored

        # Double check that we can load data from mongo even if we cannot make it
        self.st.context_config['forbid_creation_of'] = self.all_targets
        peaks = self.st.get_array(self.test_run_ids, self.all_targets[-1])
        assert len(peaks)
        runs = self.st.select_runs(available=self.all_targets)
        assert len(runs) == len(self.test_run_ids)

        # Insert a new run number and check that it's not marked as available
        self.database[self.collection_name].insert_one(_rundoc_format(3))
        self.st.runs = None  # Reset
        all_runs = self.st.select_runs()
        available_runs = self.st.select_runs(available=self.all_targets)
        assert len(available_runs) == len(self.test_run_ids)
        assert len(all_runs) == len(self.test_run_ids) + 1

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    def test_lineage_changes(self):
        st = strax.Context(register=[Records, Peaks],
                           storage=[self.rundb_sf],
                           use_per_run_defaults=True,
                           )
        lineages = [st.key_for(r, 'peaks').lineage_hash for r in self.test_run_ids]
        assert len(set(lineages)) > 1
        with self.assertRaises(ValueError):
            # Lineage changing per run is not allowed!
            st.select_runs(available='peaks')


def _rundoc_format(run_id):
    start = datetime.datetime.fromtimestamp(0) + datetime.timedelta(days=int(run_id))
    end = start + datetime.timedelta(days=1)
    doc = {
        'comments': [{'comment': 'some testdoc',
                      'date': start,
                      'user': 'master user'}],
        'data': [],
        'detectors': ['tpc'],

        'mode': 'test',
        'number': int(run_id),
        'source': 'none',
        'start': start,
        'end': end,
        'user': 'master user'}
    return doc



import numpy as np
import straxen
import tempfile
import os
import unittest
import shutil
import uuid

test_run_id_1T = '180423_1021'


class TestBasics(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        temp_folder = uuid.uuid4().hex
        # Keep one temp dir because we don't want to download the data every time.
        cls.tempdir = os.path.join(tempfile.gettempdir(), temp_folder)
        assert not os.path.exists(cls.tempdir)

        print("Downloading test data (if needed)")
        st = straxen.contexts.demo()
        cls.run_id = test_run_id_1T
        cls.st = st

    @classmethod
    def tearDownClass(cls):
        # Make sure to only cleanup this dir after we have done all the tests
        if os.path.exists(cls.tempdir):
            shutil.rmtree(cls.tempdir)

    def test_run_selection(self):
        st = self.st
        # Ignore strax-internal warnings
        st.set_context_config({'free_options': tuple(st.config.keys())})

        run_df = st.select_runs(available='raw_records')
        print(run_df)
        run_id = run_df.iloc[0]['name']
        assert run_id == test_run_id_1T

    def test_processing(self):
        df = self.st.get_df(self.run_id, 'event_info')

        assert len(df) > 0
        assert 'cs1' in df.columns
        assert df['cs1'].sum() > 0
        assert not np.all(np.isnan(df['x'].values))

    def test_event_info_double(self):
        df = self.st.get_df(self.run_id, 'event_info_double')
        assert 'cs2_a' in df.columns
        assert df['cs2_a'].sum() > 0
        assert len(df) > 0

    def test_get_livetime_sec(self):
        st = self.st
        events = st.get_array(self.run_id, 'events')
        straxen.get_livetime_sec(st, test_run_id_1T, things=events)

    def test_mini_analysis(self):
        @straxen.mini_analysis(requires=('raw_records',))
        def count_rr(raw_records):
            return len(raw_records)

        n = self.st.count_rr(self.run_id)
        assert n > 100

    @staticmethod
    def _extract_latest_comment(context,
                                test_for_target='raw_records',
                                **context_kwargs,
                                ):
        if context == 'xenonnt_online' and not straxen.utilix_is_configured():
            return
        st = getattr(straxen.contexts, context)(**context_kwargs)
        assert hasattr(st, 'extract_latest_comment'), "extract_latest_comment not added to context?"
        st.extract_latest_comment()
        assert st.runs is not None, "No registry build?"
        assert 'comments' in st.runs.keys()
        st.select_runs(available=test_for_target)
        if context == 'demo':
            assert len(st.runs)
        assert f'{test_for_target}_available' in st.runs.keys()

    def test_extract_latest_comment_nt(self, **opt):
        """Run the test for nt (but only 2000 runs"""
        self._extract_latest_comment(context='xenonnt_online',
                                     _minimum_run_number=10_000,
                                     _maximum_run_number=12_000,
                                     **opt)

    def test_extract_latest_comment_demo(self):
        self._extract_latest_comment(context='demo')

    def test_extract_latest_comment_lone_hits(self):
        """Run the test for some target that is not in the default availability check"""
        self.test_extract_latest_comment_nt(test_for_target='lone_hits')



"""
Test certain interactions with the runsdatabase.
NB! this only works if one has access to the database. This does not
work e.g. on travis jobs and therefore the tests failing locally will
not show up in Pull Requests.
"""
import straxen
import os
import unittest


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_select_runs(check_n_runs=2):
    """
    Test (if we have a connection) if we can perform strax.select_runs
        on the last two runs in the runs collection

    :param check_n_runs: int, the number of runs we want to check
    """
    assert check_n_runs >= 1
    st = straxen.contexts.xenonnt_online(use_rucio=False)
    run_col = st.storage[0].collection

    # Find the latest run in the runs collection
    last_run = run_col.find_one(projection={'number': 1},
                                sort=[('number', -1)]
                                ).get('number')

    # Set this number as the minimum run number. This limits the
    # amount of documents checked and therefore keeps the test short.
    st.storage[0].minimum_run_number = int(last_run) - (check_n_runs - 1)
    st.select_runs()


@unittest.skipIf(not straxen.utilix_is_configured(), "Cannot download because utilix is not configured")
def test_downloader():
    """Test if we can download a small file from the downloader"""
    downloader = straxen.MongoDownloader()
    path = downloader.download_single('to_pe_nt.npy')
    assert os.path.exists(path)


def _patch_om_init(take_only):
    """
    temp patch since om = straxen.OnlineMonitor() does not work with utilix
    """
    header = 'RunDB'
    user = straxen.uconfig.get(header, 'pymongo_user')
    pwd = straxen.uconfig.get(header, 'pymongo_password')
    url = straxen.uconfig.get(header, 'pymongo_url').split(',')[-1]
    uri = f"mongodb://{user}:{pwd}@{url}"
    return straxen.OnlineMonitor(uri=uri, take_only=take_only)


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_online_monitor(target='online_peak_monitor', max_tries=3):
    """
    See if we can get some data from the online monitor before max_tries

    :param target: target to test getting from the online monitor
    :param max_tries: number of queries max allowed to get a non-failing
        run
    """
    st = straxen.contexts.xenonnt_online(use_rucio=False)
    straxen.get_mongo_uri()
    om = _patch_om_init(target)
    st.storage = [om]
    max_run = None
    for i in range(max_tries):
        query = {'provides_meta': True, 'data_type': target}
        if max_run is not None:
            # One run failed before, lets try a more recent one.
            query.update({'number': {"$gt": int(max_run)}})
        some_run = om.db[om.col_name].find_one(query,
                                               projection={'number': 1,
                                                           'metadata': 1,
                                                           'lineage_hash': 1,
                                                           })
        if some_run.get('lineage_hash', False):
            if some_run['lineage_hash'] != st.key_for("0", target).lineage_hash:
                # We are doing a new release, therefore there is no
                # matching data. This makes sense.
                return
        if some_run is None or some_run.get('number', None) is None:
            print(f'Found None')
            continue
        elif 'exception' in some_run.get('metadata', {}):
            # Did find a run, but it is bad, we need to find another one
            print(f'Found {some_run.get("number", "No number")} with errors')
            max_run = some_run.get("number", -1)
            continue
        else:
            # Correctly written
            run_id = f'{some_run["number"]:06}'
            break
    else:
        raise FileNotFoundError(f'No non-failing {target} found in the online '
                                f'monitor after {max_tries}. Looked for:\n'
                                f'{st.key_for("0", target)}')
    st.get_array(run_id, target, seconds_range=(0, 1), allow_incomplete=True)



import urllib
import requests

import pandas as pd
import numba
import numpy as np

import strax
import straxen

from datetime import datetime
from datetime import timedelta
import time
import pytz

import getpass
import warnings
from configparser import NoOptionError
import sys

if any('jupyter' in arg for arg in sys.argv):
    # In some cases we are not using any notebooks,
    # Taken from 44952863 on stack overflow thanks!
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm

export, __all__ = strax.exporter()


@export
class SCADAInterface:

    def __init__(self, context=None, use_progress_bar=True):
        """
        Interface to excess the XENONnT slow control data via python.

        :param context: Context you are using e.g. st. This is needed
            if you would like to query data via run_ids.
        :param use_progress_bar: Use a progress bar in the Scada interface
        """
        self.we_are_straxen = False
        self._token_expire_time = None
        self._token = None
        self.pmt_file_found = True
        try:
            self.SCLogin_url = straxen.uconfig.get('scada', 'sclogin_url')
            self.SCData_URL = straxen.uconfig.get('scada', 'scdata_url')
            self.SCLastValue_URL = straxen.uconfig.get('scada', 'sclastvalue_url')

        except ValueError as e:
            raise ValueError(f'Cannot load SCADA information, from your xenon'
                             ' config. SCADAInterface cannot be used.') from e

        try:
            # Load parameters from the database.
            self.pmt_file = straxen.get_resource('PMTmap_SCADA.json', fmt='json')
        except FileNotFoundError:
            warnings.warn('Cannot find PMT map, "find_pmt_names" cannot be used.')
            self.pmt_file_found = False

        # Use a tqdm progress bar if requested. If a user does not want
        # a progress bar, just wrap it by a tuple
        self._use_progress_bar = use_progress_bar
        self.context = context

        self.we_are_straxen = True
        self.get_new_token()

    def get_scada_values(self,
                         parameters,
                         start=None,
                         end=None,
                         run_id=None,
                         query_type_lab=True,
                         time_selection_kwargs=None,
                         fill_gaps=None,
                         filling_kwargs=None,
                         down_sampling=False,
                         every_nth_value=1):
        """
        Function which returns XENONnT slow control values for a given
        set of parameters and time range.

        The time range can be either defined by a start and end time or
        via the run_id, target and context.

        :param parameters: dictionary containing the names of the
            requested scada-parameters. The keys are used as identifier
            of the parameters in the returned pandas.DataFrame.
        :param start: int representing the start time of the interval
            in ns unix time.
        :param end: same as start but as end.
        :param run_id: Id of the run. Can also be specified as a list or
            tuple of run ids. In this case we will return the time
            range lasting between the start of the first and endtime
            of the second run.
        :param query_type_lab: Mode on how to query data from the historians.
            Can be either False to get raw data or True (default) to get
            data which was interpolated by historian. Useful if large
            time ranges have to be queried.
        :param time_selection_kwargs: Keyword arguments taken by
            st.to_absolute_time_range(). Default: {"full_range": True}
        :param fill_gaps: Decides how to fill gaps in which no data was
            recorded. Only needed for query_type_lab=False. Can be either
            None, "interpolation" or "forwardfill".None keeps the gaps
            (default), "interpolation" uses pandas.interpolate and
            "forwardfill" pandas.ffill. See
            https://pandas.pydata.org/docs/ for more information. You
            can change the filling options of the methods with the
            filling_kwargs.
        :param filling_kwargs: Kwargs applied to pandas .ffill() or
            .interpolate(). Only needed for query_type_lab=False.
        :param down_sampling: Boolean which indicates whether to
            donw_sample result or to apply average. The averaging
            is deactivated in case of interpolated data. Only needed
            for query_type_lab=False.
        :param every_nth_value: Defines over how many values we compute
            the average or the nth sample in case we down sample the
            data. In case query_type_lab=True every nth second is
            returned.
        :return: pandas.DataFrame containing the data of the specified
            parameters.
        """

        if not filling_kwargs:
            filling_kwargs = {}

        if not isinstance(parameters, dict):
            mes = 'The argument "parameters" has to be specified as a dict.'
            raise ValueError(mes)

        start, end, now = self._get_and_check_start_end(run_id,
                                                        start,
                                                        end,
                                                        time_selection_kwargs
                                                        )

        _fill_gaps = [None, 'None', 'interpolation', 'forwardfill']
        if fill_gaps not in _fill_gaps:
            raise ValueError(f'Wrong argument for "fill_gaps", must be either {_fill_gaps}.'
                             f' You specified "{fill_gaps}"')

        if not self._token:
            # User has not asked for a token yet:
            self._get_token()

        # Check if token will expire soon, if so renew the token before we query
        # the parameters:
        hrs, mins = self._token_expires_in()
        if hrs == 0 and mins < 30:
            print('Your token will expire in less than 30 min please get first a new one:')
            self._get_token()

        # Now loop over specified parameters and get the values for those.
        iterator = enumerate(parameters.items())
        if self._use_progress_bar:
            # wrap using progress bar
            iterator = tqdm(iterator, total=len(parameters), desc='Load parameters')

        for ind, (k, p) in iterator:
            try:
                temp_df = self._query_single_parameter(start, end,
                                                       k, p,
                                                       every_nth_value=every_nth_value,
                                                       fill_gaps=fill_gaps,
                                                       filling_kwargs=filling_kwargs,
                                                       down_sampling=down_sampling,
                                                       query_type_lab=query_type_lab)
                
                if ind:
                    m = np.all(df.loc[:, 'time'] == temp_df.loc[:, 'time'])
                    if ind and not m:
                        raise ValueError('This is odd somehow the time stamps for the query of'
                                         f' {p} does not match the previous timestamps.')
            except ValueError as e:
                warnings.warn(f'Was not able to load parameters for "{k}". The reason was: "{e}".'
                               f'Continue without {k}.')
                temp_df = pd.DataFrame(columns=(k,))
            
            if ind:
                df = pd.concat((df, temp_df[k]), axis=1)
            else:
                df = temp_df
                
        # Adding timezone information and rename index:
        df.set_index('time', inplace=True)
        df = df.tz_localize(tz='UTC')
        df.index.rename('time UTC', inplace=True)

        if (end // 10**9) > now.astype(np.int64):
            df.loc[now:, :] = np.nan

        return df

    def _get_and_check_start_end(self, run_id, start, end, time_selection_kwargs):
        """
        Helper function which clusters all time related checks and reduces complexity
        of get_scada_values.
        """
        if not time_selection_kwargs:
            time_selection_kwargs = {'full_range': True}

        if np.all((run_id, self.context)):
            # User specified a valid context and run_id, so get the start
            # and end time for our query:
            if isinstance(run_id, (list, tuple)):
                run_id = np.sort(run_id)  # Do not trust the user's
                start, _ = self.context.to_absolute_time_range(run_id[0], **time_selection_kwargs)
                _, end = self.context.to_absolute_time_range(run_id[-1], **time_selection_kwargs)
            else:
                start, end = self.context.to_absolute_time_range(run_id, **time_selection_kwargs)
        elif run_id:
            mes = ('You are trying to query slow control data via run_ids'
                   ' but you have not specified the context you are '
                   'working with. Please set the context either via '
                   '.context = YOURCONTEXT, or when initializing the '
                   'interface.')
            raise ValueError(mes)

        if not np.all((start, end)):
            # User has not specified any valid start and end time
            mes = ('You have to specify either a run_id and context.'
                   ' E.g. call get_scada_values(parameters, run_id=run)'
                   ' or you have to specify a valid start and end time '
                   'in utc unix time ns.')
            raise ValueError(mes)

        if end < start:
            raise ValueError('You specified an endtime which is smaller '
                             'than the start time.')

        if (np.log10(start) < 18) or (np.log10(end) < 18):
            raise ValueError('Expected the time to be in ns unix time (number with 19 digits or more).'
                             ' Have you specified the time maybe in seconds or micro-seconds?')

        now = np.datetime64('now')
        if (end // 10**9) > now.astype(np.int64):
            mes = ('You are asking for an endtime which is in the future,'
                   ' I may be written by a physicist, but I am neither self-'
                   'aware nor can I predict the future like they can. You '
                   f'asked for the endtime: {end // 10**9} but current utc '
                   f'time is {now.astype(np.int64)}. I will return for the values for the '
                   'corresponding times as nans instead.')
            warnings.warn(mes)

        # Chop start/end time if precision is higher then seconds level.
        start = (start//10**9)*10**9
        end = (end//10**9)*10**9

        return int(start), int(end), now

    def _query_single_parameter(self,
                                start,
                                end,
                                parameter_key,
                                parameter_name,
                                fill_gaps,
                                filling_kwargs,
                                down_sampling,
                                query_type_lab=False,
                                every_nth_value=1):
        """
        Function to query the values of a single parameter from SCData.

        :param start: Start time in ns unix time
        :param end: End time in ns unix time
        :param parameter_key: Key to identify queried parameter in the
            DataFrame
        :param parameter_name: Parameter name in Scada/historian database.
        :param fill_gaps: Decides how to fill gaps in which no data was
            recorded. Only needed for query_type_lab=False. Can be either
            None, "interpolation" or "forwardfill".None keeps the gaps
            (default), "interpolation" uses pandas.interpolate and
            "forwardfill" pandas.ffill. See
            https://pandas.pydata.org/docs/ for more information. You
            can change the filling options of the methods with the
            filling_kwargs.
        :param filling_kwargs: Keyword arguments forwarded to pandas.ffill
            or pandas.interpolate.
        :param every_nth_value: Defines over how many values we compute
            the average or the nthed sample in case we down sample the
            data.

        :returns: DataFrame with a time and parameter_key column.
        """
        if every_nth_value < 1:
            mes = ("SCADA takes only values every second. Cannot ask for a"
                   " higher sampling rate than one value per second. However"
                   f" you asked for one value every {every_nth_value} seconds.")
            raise ValueError(mes)
        if not isinstance(every_nth_value, int):
            raise ValueError('"value_every_seconds" must be an int!')

        # First we have to create an array where we can fill values:
        if query_type_lab:
            # In the lab case we get interpolated data without nans so the df can be set
            # accordingly.
            seconds = np.arange(start, end + 1, 10**9 * every_nth_value)
        else:
            seconds = np.arange(start, end + 1, 10**9)  # +1 to make sure endtime is included

        df = pd.DataFrame()
        df.loc[:, 'time'] = seconds
        df['time'] = df['time'].astype('<M8[ns]')
        df.set_index('time', inplace=True)

        # Init parameter query:
        query = {'name': parameter_name}

        if not query_type_lab:
            # Check if first value is in requested range:
            # This is only needed in case of raw data since here it can 
            # happen that the user queries a range without any data.
            temp_df = self._query(query,
                                  self.SCLastValue_URL,
                                  end=(start // 10**9) + 1)  # +1 since it is end before exclusive

            # Store value as first value in our df
            df.loc[df.index.values[0], parameter_key] = temp_df['value'][0]
            offset = 1
        else:
            offset = 0

        one_year_in_ns = int(24*3600*360*10**9)
        starts = np.arange(start+offset, end, one_year_in_ns)
        if len(starts):
            ends = starts + one_year_in_ns
            ends = np.clip(ends, a_max=end, a_min=0)
        else:
            ends = np.array([end])
            starts = np.array([start])

        for start_query, end_query in zip(starts, ends):
            self._query_data_per_year(parameter_key,
                                      query,
                                      start_query,
                                      end_query,
                                      query_type_lab,
                                      every_nth_value,
                                      df,
                                      )

        # Let user decided whether to ffill, interpolate or keep gaps:
        if fill_gaps == 'interpolation':
            df.interpolate(**filling_kwargs, inplace=True)

        if fill_gaps == 'forwardfill':
            # Now fill values in between like Scada would do:
            df.ffill(**filling_kwargs, inplace=True)

        # Step 4. Down-sample data if asked for:
        df.reset_index(inplace=True)
        if every_nth_value > 1 and not query_type_lab:
            # If the user asks for down sampling do so, but only for 
            # raw_data, lab query type is already interpolated and down sampled
            # by the historian. 
            if down_sampling:
                df = df[::every_nth_value]
            else:
                nt, nv = _average_scada(df['time'].astype(np.int64).values,
                                        df[parameter_key].values,
                                        every_nth_value)
                df = pd.DataFrame()
                df['time'] = nt.astype('<M8[ns]')
                df[parameter_key] = nv

        return df

    def _query_data_per_year(self, 
                             parameter_name,
                             query,
                             start,
                             end,
                             query_type_lab,
                             seconds_interval,
                             result_dataframe,
                             ):
        """
        The SCADA API cannot handle query ranges lasting longer than
        one year. So in case the user specifies a longer time range
        we have to chunk the time requests in steps of years.

        Updates the resulting dataframe in place.
        """
        ntries = 0
        # This corresponds to a bit more than one year assuming 1 value per second:
        max_tries = 1000
        while ntries < max_tries:
            # Although we step the query already in years we also have to
            # do the query in a whole loop as we can only query 35000
            # data points at any given time, however it is not possible
            # to know the sampling rate of the queried parameter apriori.
            temp_df = self._query(query,
                                  self.SCData_URL,
                                  start=(start // 10**9),
                                  end=(end // 10**9),
                                  query_type_lab=query_type_lab,
                                  seconds_interval=seconds_interval,
                                  raise_error_message=False,  # No valid value in query range...
                                  )  # +1 since it is end before exclusive
            if temp_df.empty:
                # In case WebInterface does not return any data, e.g. if query range too small
                break

            times = (temp_df['timestampseconds'].values * 10**9).astype('<M8[ns]')
            result_dataframe.loc[times, parameter_name] = temp_df.loc[:, 'value'].values
            endtime = temp_df['timestampseconds'].values[-1].astype(np.int64)
            start = endtime  # Next query should start at the last time seen.

            ntries += 1
            if not (len(temp_df) == 35000 and endtime != end // 10**9):
                # Max query are 35000 values, if end is reached the
                # length of the dataframe is either smaller or the last
                # time value is equivalent to queried range.
                break

        return result_dataframe

    def _query(self,
               query,
               api,
               start=None,
               end=None,
               query_type_lab=False,
               seconds_interval=None,
               raise_error_message=True):
        """
        Helper to reduce code. Asks for data and returns result. Raises error
        if api returns error.
        """
        if start:
            query["StartDateUnix"] = start
        if end:
            query['EndDateUnix'] = end

        if query_type_lab:
            query['QueryType'] = 'lab'
            query['interval'] = seconds_interval
        else:
            query['QueryType'] = 'rawbytime'
            query.pop('interval', None)  # Interval only works with lab

        # Configure query url
        query_url = urllib.parse.urlencode(query)
        self._query_url = api + query_url
        # Security check if url is a real url and not something like file://
        if not self._query_url.lower().startswith('https'):
            raise ValueError('The query URL should start with https! '
                             f'Current URL: {self._query_url}')

        response = requests.get(self._query_url,
                                headers={'Authorization': self._token})
        if response.status_code == 401:
            # Invalid token so we have to get a new one,
            # this should actually never happen, but you never know...
            print('Your token is invalid. It may have expired please get a new one:')
            # If the user puts in the wrong credentials the query will fail.
            self._get_token()
            response = requests.get(self._query_url,
                                    headers={'Authorization': self._token})

        if response.status_code != 200:
            # Check if we get any status code different from 200 == ok
            # If yes raise the corresponding status:
            response.raise_for_status()

        # Read database response and check if query was valid:
        values = response.json()

        temp_df = pd.DataFrame(columns=('timestampseconds', 'value'))
        if isinstance(values, dict) and raise_error_message:
            # Not valid, why:
            query_status = values['status']
            query_message = values['message']
            raise ValueError(f'SCADAapi has not returned values for the '
                             f'parameter "{query["name"]}". It returned the '
                             f'status "{query_status}" with the message "{query_message}".')
        if isinstance(values, list):
            # Valid so return dataframe
            temp_df = pd.DataFrame(values)
        return temp_df

    def find_scada_parameter(self):
        raise NotImplementedError('Feature not implemented yet.')

    def find_pmt_names(self, pmts=None, hv=True, current=False):
        """
        Function which returns a list of PMT parameter names to be
        called in SCADAInterface.get_scada_values. The names refer to
        the high voltage of the PMTs, not their current.

        Thanks to Hagar and Giovanni who provided the file.

        :param pmts: Optional parameter to specify which PMT parameters
            should be returned. Can be either a list or array of channels
            or just a single one.
        :param hv: Bool if true names of high voltage channels are
            returned.
        :param current: Bool if true names for the current channels are
            returned.
        :return: dictionary containing short names as keys and scada
            parameter names as values.
        """
        if not self.pmt_file_found:
            raise ValueError('json file containing the PMT information was not found. '
                             '"find_pmt_names" cannot be used.')

        if not (hv or current):
            raise ValueError('Either one "hv" or "current" must be true.')

        if isinstance(pmts, np.ndarray):
            # convert to a simple list since otherwise we get ambiguous errors
            pmts = list(pmts)

        if not hasattr(pmts, '__iter__'):
            # If single PMT convert it to itterable
            pmts = [pmts]

        # Getting parameter names for all PMTs:
        # The file contains the names for the HV channels
        if pmts:
            pmts_v = {k: v for k, v in self.pmt_file.items() if int(k[3:]) in pmts}
        else:
            pmts_v = self.pmt_file

        res = {}
        # Now get all relevant names:
        for key, value in pmts_v.items():
            if hv:
                res[key + '_HV'] = value
            if current:
                res[key + '_I'] = value[:-4] + 'IMON'

        return res

    def get_new_token(self):
        """
        Function to renew the token of the current session.
        """
        self._get_token()

    def _get_token(self):
        """
        Function which asks for user credentials to receive a personalized
        security token. The token is required to query any data from the
        slow control historians.
        """
        if not self.we_are_straxen:
            username, password =  self._ask_for_credentials()
        else:
            try:
                username = straxen.uconfig.get('scada', 'straxen_username')
                password = straxen.uconfig.get('scada', 'straxen_password')
            except (AttributeError, NoOptionError):
                # If section does not exist Fall back to user credentials
                username, password = self._ask_for_credentials()

        login_query = {'username': username,
                       'password': password,
                       }
        res = requests.post(self.SCLogin_url,
                            data=login_query)

        res = res.json()
        if 'token' not in res.keys():
            raise ValueError('Cannot get security token from Slow Control web API. '
                             f'API returned the following reason: {res["Message"]}')

        self._token = res['token']
        toke_start_time = datetime.now(tz=pytz.timezone('utc'))
        hours_added = timedelta(hours=3)
        self._token_expire_time = toke_start_time + hours_added
        print('Received token, the token is valid for 3 hrs.\n',
              f'from {toke_start_time.strftime("%d.%m. %H:%M:%S")} UTC\n',
              f'till {self._token_expire_time.strftime("%d.%m. %H:%M:%S")} UTC\n'
              'We will automatically refresh the token for you :). '
              'Have a nice day and a fruitful analysis!'
              )

    @staticmethod
    def _ask_for_credentials():
        print('Please, enter your Xe1TViewer/SCADA credentials:')
        time.sleep(1)
        username = getpass.getpass('Xenon Username: ')
        password = getpass.getpass('Xenon Password: ')
        return username, password

    def token_expires_in(self):
        """
        Function which displays how long until the current token expires.
        """
        if self._token_expire_time:
            print(f'The current token expires at {self._token_expire_time.strftime("%d.%m. %H:%M:%S")} UTC')
            hrs, mins = self._token_expires_in()
            print(f'Which is in {hrs} h and {mins} min.')
        else:
            raise ValueError('You do not have any valid token yet. Please call '
                             '"get_new_token" first".')

    def _token_expires_in(self):
        """
        Computes hrs and minutes until token expires.
        """
        now = datetime.now(tz=pytz.timezone('utc'))
        dt = (self._token_expire_time - now).seconds  # time delta in seconds
        hrs = dt // 3600
        mins = dt % 3600 // 60
        return hrs, mins


@export
def convert_time_zone(df, tz):
    """
    Function which converts the current time zone of a given
    pd.DataFrame into another timezone.

    :param df: pandas.DataFrame containing the Data. Index must be a
        datetime object with time zone information.
    :param tz: str representing the timezone the index should be
        converted to. See the notes for more information.
    :return: pandas.DataFrame with converted time index.

    Notes:
        1. ) The input pandas.DataFrame must be indexed via datetime
        objects which are timezone aware.

        2.)  You can find a complete list of available timezones via:
        ```
        import pytz
        pytz.all_timezones
        ```
        You can also specify 'strax' as timezone which will convert the
        time index into a 'strax time' equivalent.
        The default timezone of strax is UTC.
    """
    if tz == 'strax':
        df = df.tz_convert(tz='UTC')
        df.index = df.index.astype(np.int64)
        df.index.rename(f'time strax', inplace=True)
    else:
        df = df.tz_convert(tz=tz)
        df.index.rename(f'time {tz}', inplace=True)
    return df


@numba.njit
def _average_scada(times, values, nvalues):
    """
    Function which down samples scada values.

    :param times: Unix times of the data points.
    :param values: Corresponding sensor value
    :param nvalues: Number of samples we average over.
    :return: new time values and
    """
    if len(times) % nvalues:
        n_samples = (len(times) // nvalues) - 1
    else:
        n_samples = (len(times) // nvalues)
    res = np.zeros(n_samples, dtype=np.float32)
    new_times = np.zeros(n_samples, dtype=np.int64)
    for ind in range(n_samples):
        res[ind] = np.mean(values[ind * nvalues:(ind + 1) * nvalues])
        new_times[ind] = np.mean(times[ind * nvalues:(ind + 1) * nvalues])

    return new_times, res
